import numpy as np

try:
    import tensorflow.compat.v1 as tf
except ImportError as error:
    from logging import warning
    warning("{}: AdversarialDebiasing will be unavailable. To install, run:\n"
            "pip install 'aif360[AdversarialDebiasing]'".format(error))

from aif360.algorithms import Transformer


class AdversarialDebiasing(Transformer):
    """Adversarial debiasing is an in-processing technique that learns a
    classifier to maximize prediction accuracy and simultaneously reduce an
    adversary's ability to determine the protected attribute from the
    predictions [5]_. This approach leads to a fair classifier as the
    predictions cannot carry any group discrimination information that the
    adversary can exploit.

    References:
        .. [5] B. H. Zhang, B. Lemoine, and M. Mitchell, "Mitigating Unwanted
           Biases with Adversarial Learning," AAAI/ACM Conference on Artificial
           Intelligence, Ethics, and Society, 2018.
    """

    def __init__(self,
                 unprivileged_groups,
                 privileged_groups,
                 scope_name,
                 sess,
                 seed=None,
                 adversary_loss_weight=0.1,
                 num_epochs=50,
                 batch_size=128,
                 classifier_num_hidden_units=200,
                 debias=True):
        """
        Args:
            unprivileged_groups (tuple): Representation for unprivileged groups
            privileged_groups (tuple): Representation for privileged groups
            scope_name (str): scope name for the tenforflow variables
            sess (tf.Session): tensorflow session
            seed (int, optional): Seed to make `predict` repeatable.
            adversary_loss_weight (float, optional): Hyperparameter that chooses
                the strength of the adversarial loss.
            num_epochs (int, optional): Number of training epochs.
            batch_size (int, optional): Batch size.
            classifier_num_hidden_units (int, optional): Number of hidden units
                in the classifier model.
            debias (bool, optional): Learn a classifier with or without
                debiasing.
        """
        super(AdversarialDebiasing, self).__init__(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups)

        self.scope_name = scope_name
        self.seed = seed

        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups
        if len(self.unprivileged_groups) > 1 or len(self.privileged_groups) > 1:
            raise ValueError("Only one unprivileged_group or privileged_group supported.")
        self.protected_attribute_name = list(self.unprivileged_groups[0].keys())[0]

        self.sess = sess
        self.adversary_loss_weight = adversary_loss_weight
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.classifier_num_hidden_units = classifier_num_hidden_units
        self.debias = debias

        self.features_dim = None
        self.features_ph = None
        self.protected_attributes_ph = None
        self.true_labels_ph = None
        self.pred_labels = None

    def _classifier_model(self, features, features_dim, keep_prob):
        """Compute the classifier predictions for the outcome variable.
        """
        with tf.variable_scope("classifier_model"):
            W1 = tf.get_variable('W1', [features_dim, self.classifier_num_hidden_units],
                                  initializer=tf.initializers.glorot_uniform(seed=self.seed1))
            b1 = tf.Variable(tf.zeros(shape=[self.classifier_num_hidden_units]), name='b1')

            h1 = tf.nn.relu(tf.matmul(features, W1) + b1)
            h1 = tf.nn.dropout(h1, keep_prob=keep_prob, seed=self.seed2)

            W2 = tf.get_variable('W2', [self.classifier_num_hidden_units, 1],
                                 initializer=tf.initializers.glorot_uniform(seed=self.seed3))
            b2 = tf.Variable(tf.zeros(shape=[1]), name='b2')

            pred_logit = tf.matmul(h1, W2) + b2
            pred_label = tf.sigmoid(pred_logit)

        return pred_label, pred_logit

    def _adversary_model(self, pred_logits, true_labels):
        """Compute the adversary predictions for the protected attribute.
        """
        with tf.variable_scope("adversary_model"):
            c = tf.get_variable('c', initializer=tf.constant(1.0))
            s = tf.sigmoid((1 + tf.abs(c)) * pred_logits)

            W2 = tf.get_variable('W2', [3, 1],
                                 initializer=tf.initializers.glorot_uniform(seed=self.seed4))
            b2 = tf.Variable(tf.zeros(shape=[1]), name='b2')

            pred_protected_attribute_logit = tf.matmul(tf.concat([s, s * true_labels, s * (1.0 - true_labels)], axis=1), W2) + b2
            pred_protected_attribute_label = tf.sigmoid(pred_protected_attribute_logit)

        return pred_protected_attribute_label, pred_protected_attribute_logit

    def fit(self, dataset):
        """Compute the model parameters of the fair classifier using gradient
        descent.

        Args:
            dataset (BinaryLabelDataset): Dataset containing true labels.

        Returns:
            AdversarialDebiasing: Returns self.
        """
        if tf.executing_eagerly():
            raise RuntimeError("AdversarialDebiasing does not work in eager "
                    "execution mode. To fix, add `tf.disable_eager_execution()`"
                    " to the top of the calling script.")

        if self.seed is not None:
            np.random.seed(self.seed)
        ii32 = np.iinfo(np.int32)
        self.seed1, self.seed2, self.seed3, self.seed4 = np.random.randint(ii32.min, ii32.max, size=4)

        # Map the dataset labels to 0 and 1.
        temp_labels = dataset.labels.copy()

        temp_labels[(dataset.labels == dataset.favorable_label).ravel(),0] = 1.0
        temp_labels[(dataset.labels == dataset.unfavorable_label).ravel(),0] = 0.0

        with tf.variable_scope(self.scope_name):
            num_train_samples, self.features_dim = np.shape(dataset.features)

            # Setup placeholders
            self.features_ph = tf.placeholder(tf.float32, shape=[None, self.features_dim])
            self.protected_attributes_ph = tf.placeholder(tf.float32, shape=[None,1])
            self.true_labels_ph = tf.placeholder(tf.float32, shape=[None,1])
            self.keep_prob = tf.placeholder(tf.float32)

            # Obtain classifier predictions and classifier loss
            self.pred_labels, pred_logits = self._classifier_model(self.features_ph, self.features_dim, self.keep_prob)
            pred_labels_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=self.true_labels_ph, logits=pred_logits))

            if self.debias:
                # Obtain adversary predictions and adversary loss
                pred_protected_attributes_labels, pred_protected_attributes_logits = self._adversary_model(pred_logits, self.true_labels_ph)
                pred_protected_attributes_loss = tf.reduce_mean(
                    tf.nn.sigmoid_cross_entropy_with_logits(labels=self.protected_attributes_ph, logits=pred_protected_attributes_logits))

            # Setup optimizers with learning rates
            global_step = tf.Variable(0, trainable=False)
            starter_learning_rate = 0.001
            learning_rate = tf.train.exponential_decay(starter_learning_rate, global_step,
                                                       1000, 0.96, staircase=True)
            classifier_opt = tf.train.AdamOptimizer(learning_rate)
            if self.debias:
                adversary_opt = tf.train.AdamOptimizer(learning_rate)

            classifier_vars = [var for var in tf.trainable_variables(scope=self.scope_name) if 'classifier_model' in var.name]
            if self.debias:
                adversary_vars = [var for var in tf.trainable_variables(scope=self.scope_name) if 'adversary_model' in var.name]
                # Update classifier parameters
                adversary_grads = {var: grad for (grad, var) in adversary_opt.compute_gradients(pred_protected_attributes_loss,
                                                                                      var_list=classifier_vars)}
            normalize = lambda x: x / (tf.norm(x) + np.finfo(np.float32).tiny)

            classifier_grads = []
            for (grad,var) in classifier_opt.compute_gradients(pred_labels_loss, var_list=classifier_vars):
                if self.debias:
                    unit_adversary_grad = normalize(adversary_grads[var])
                    grad -= tf.reduce_sum(grad * unit_adversary_grad) * unit_adversary_grad
                    grad -= self.adversary_loss_weight * adversary_grads[var]
                classifier_grads.append((grad, var))
            classifier_minimizer = classifier_opt.apply_gradients(classifier_grads, global_step=global_step)

            if self.debias:
                # Update adversary parameters
                with tf.control_dependencies([classifier_minimizer]):
                    adversary_minimizer = adversary_opt.minimize(pred_protected_attributes_loss, var_list=adversary_vars)#, global_step=global_step)

            self.sess.run(tf.global_variables_initializer())
            self.sess.run(tf.local_variables_initializer())

            # Begin training
            for epoch in range(self.num_epochs):
                shuffled_ids = np.random.choice(num_train_samples, num_train_samples, replace=False)
                for i in range(num_train_samples//self.batch_size):
                    batch_ids = shuffled_ids[self.batch_size*i: self.batch_size*(i+1)]
                    batch_features = dataset.features[batch_ids]
                    batch_labels = np.reshape(temp_labels[batch_ids], [-1,1])
                    batch_protected_attributes = np.reshape(dataset.protected_attributes[batch_ids][:,
                                                 dataset.protected_attribute_names.index(self.protected_attribute_name)], [-1,1])

                    batch_feed_dict = {self.features_ph: batch_features,
                                       self.true_labels_ph: batch_labels,
                                       self.protected_attributes_ph: batch_protected_attributes,
                                       self.keep_prob: 0.8}
                    if self.debias:
                        _, _, pred_labels_loss_value, pred_protected_attributes_loss_vale = self.sess.run([classifier_minimizer,
                                       adversary_minimizer,
                                       pred_labels_loss,
                                       pred_protected_attributes_loss], feed_dict=batch_feed_dict)
                        if i % 200 == 0:
                            print("epoch %d; iter: %d; batch classifier loss: %f; batch adversarial loss: %f" % (epoch, i, pred_labels_loss_value,
                                                                                     pred_protected_attributes_loss_vale))
                    else:
                        _, pred_labels_loss_value = self.sess.run(
                            [classifier_minimizer,
                             pred_labels_loss], feed_dict=batch_feed_dict)
                        if i % 200 == 0:
                            print("epoch %d; iter: %d; batch classifier loss: %f" % (
                            epoch, i, pred_labels_loss_value))
        return self

    def predict(self, dataset):
        """Obtain the predictions for the provided dataset using the fair
        classifier learned.

        Args:
            dataset (BinaryLabelDataset): Dataset containing labels that needs
                to be transformed.
        Returns:
            dataset (BinaryLabelDataset): Transformed dataset.
        """

        if self.seed is not None:
            np.random.seed(self.seed)

        num_test_samples, _ = np.shape(dataset.features)

        samples_covered = 0
        pred_labels = []
        while samples_covered < num_test_samples:
            start = samples_covered
            end = samples_covered + self.batch_size
            if end > num_test_samples:
                end = num_test_samples
            batch_ids = np.arange(start, end)
            batch_features = dataset.features[batch_ids]
            batch_labels = np.reshape(dataset.labels[batch_ids], [-1,1])
            batch_protected_attributes = np.reshape(dataset.protected_attributes[batch_ids][:,
                                         dataset.protected_attribute_names.index(self.protected_attribute_name)], [-1,1])

            batch_feed_dict = {self.features_ph: batch_features,
                               self.true_labels_ph: batch_labels,
                               self.protected_attributes_ph: batch_protected_attributes,
                               self.keep_prob: 1.0}

            pred_labels += self.sess.run(self.pred_labels, feed_dict=batch_feed_dict)[:,0].tolist()
            samples_covered += len(batch_features)

        # Mutated, fairer dataset with new labels
        dataset_new = dataset.copy(deepcopy = True)
        dataset_new.scores = np.array(pred_labels, dtype=np.float64).reshape(-1, 1)
        dataset_new.labels = (np.array(pred_labels)>0.5).astype(np.float64).reshape(-1,1)


        # Map the dataset labels to back to their original values.
        temp_labels = dataset_new.labels.copy()

        temp_labels[(dataset_new.labels == 1.0).ravel(), 0] = dataset.favorable_label
        temp_labels[(dataset_new.labels == 0.0).ravel(), 0] = dataset.unfavorable_label

        dataset_new.labels = temp_labels.copy()

        return dataset_new



# Based on code from https://github.com/zjelveh/learning-fair-representations

import numpy as np
from scipy.spatial.distance import cdist
from scipy.special import softmax


def LFR_optim_objective(parameters, x_unprivileged, x_privileged, y_unprivileged,
                        y_privileged, k=10, A_x=0.01, A_y=0.1, A_z=0.5, print_interval=250, verbose=1):

    num_unprivileged, features_dim = x_unprivileged.shape
    num_privileged, _ = x_privileged.shape

    w = parameters[:k]
    prototypes = parameters[k:].reshape((k, features_dim))

    M_unprivileged, x_hat_unprivileged, y_hat_unprivileged = get_xhat_y_hat(prototypes, w, x_unprivileged)

    M_privileged, x_hat_privileged, y_hat_privileged = get_xhat_y_hat(prototypes, w, x_privileged)

    y_hat = np.concatenate([y_hat_unprivileged, y_hat_privileged], axis=0)
    y = np.concatenate([y_unprivileged.reshape((-1, 1)), y_privileged.reshape((-1, 1))], axis=0)

    L_x = np.mean((x_hat_unprivileged - x_unprivileged) ** 2) + np.mean((x_hat_privileged - x_privileged) ** 2)
    L_z = np.mean(abs(np.mean(M_unprivileged, axis=0) - np.mean(M_privileged, axis=0)))
    L_y = - np.mean(y * np.log(y_hat) + (1. - y) * np.log(1. - y_hat))

    total_loss = A_x * L_x + A_y * L_y + A_z * L_z

    if verbose and LFR_optim_objective.steps % print_interval == 0:
        print("step: {}, loss: {}, L_x: {},  L_y: {},  L_z: {}".format(
            LFR_optim_objective.steps, total_loss, L_x,  L_y,  L_z))
    LFR_optim_objective.steps += 1

    return total_loss


def get_xhat_y_hat(prototypes, w, x):
    M = softmax(-cdist(x, prototypes), axis=1)
    x_hat = np.matmul(M, prototypes)
    y_hat = np.clip(
        np.matmul(M, w.reshape((-1, 1))),
        np.finfo(float).eps,
        1.0 - np.finfo(float).eps
    )
    return M, x_hat, y_hat



from typing import Union

from aif360.detectors.mdss.ScoringFunctions import (
    Bernoulli,
    BerkJones,
    Gaussian,
    ScoringFunction,
    Poisson,
)
from aif360.detectors.mdss.MDSS import MDSS

import pandas as pd


def bias_scan(
    data: pd.DataFrame,
    observations: pd.Series,
    expectations: Union[pd.Series, pd.DataFrame] = None,
    favorable_value: Union[str, float] = None,
    overpredicted: bool = True,
    scoring: Union[str, ScoringFunction] = "Bernoulli",
    num_iters: int = 10,
    penalty: float = 1e-17,
    mode: str = "binary",
    **kwargs,
):
    """
    scan to find the highest scoring subset of records

    :param data (dataframe): the dataset (containing the features) the model was trained on
    :param observations (series): ground truth (correct) target values
    :param expectations (series,  dataframe, optional): pandas series estimated targets
        as returned by a model for binary, continuous and ordinal modes.
        If mode is nominal, this is a dataframe with columns containing expectations for each nominal class.
        If None, model is assumed to be a dumb model that predicts the mean of the targets
                or 1/(num of categories) for nominal mode.
    :param favorable_value(str, float, optional): Should be high or low or float if the mode in [binary, ordinal, or continuous].
            If float, value has to be minimum or maximum in the observations column. Defaults to high if None for these modes.
            Support for float left in to keep the intuition clear in binary classification tasks.
            If mode is nominal, favorable values should be one of the unique categories in the observations.
            Defaults to a one-vs-all scan if None for nominal mode.
    :param overpredicted (bool, optional): flag for group to scan for.
        True means we scan for a group whose expectations/predictions are systematically higher than observed.
        In other words, True means we scan for a group whose observeed is systematically lower than the expectations.
        False means we scan for a group whose expectations/predictions are systematically lower than observed.
        In other words, False means we scan for a group whose observed is systematically higher than the expectations.
    :param scoring (str or class): One of 'Bernoulli', 'Gaussian', 'Poisson', or 'BerkJones' or subclass of
            :class:`aif360.metrics.mdss.ScoringFunctions.ScoringFunction`.
    :param num_iters (int, optional): number of iterations (random restarts). Should be positive.
    :param penalty (float,optional): penalty term. Should be positive. The penalty term as with any regularization parameter may need to be
        tuned for ones use case. The higher the penalty, the less complex (number of features and feature values) the
        highest scoring subset that gets returned is.
    :param mode: one of ['binary', 'continuous', 'nominal', 'ordinal']. Defaults to binary.
            In nominal mode, up to 10 categories are supported by default.
            To increase this, pass in keyword argument max_nominal = integer value.

    :returns: the highest scoring subset and the score or dict of the highest scoring subset and the score for each category in nominal mode
    """
    # Ensure correct mode is passed in.
    modes = ["binary", "continuous", "nominal", "ordinal"]
    assert mode in modes, f"Expected one of {modes}, got {mode}."

    # Set correct favorable value (this tells us if higher or lower is better)
    min_val, max_val = observations.min(), observations.max()
    uniques = list(observations.unique())

    if favorable_value == 'high':
        favorable_value = max_val
    elif favorable_value == 'low':
        favorable_value = min_val
    elif favorable_value is None:
        if mode in ["binary", "ordinal", "continuous"]:
            favorable_value = max_val # Default to higher is better
        elif mode == "nominal":
            favorable_value = "flag-all" # Default to scan through all categories
            assert favorable_value in [
                "flag-all",
                *uniques,
            ], f"Expected one of {uniques}, got {favorable_value}."

    assert favorable_value in [
        min_val,
        max_val,
        "flag-all",
        *uniques,
    ], f"Favorable_value should be high, low, or one of categories {uniques}, got {favorable_value}."

    # Set appropriate direction for scanner depending on mode and overppredicted flag
    if mode in ["ordinal", "continuous"]:
        if favorable_value == max_val:
            kwargs["direction"] = "negative" if overpredicted else "positive"
        else:
            kwargs["direction"] = "positive" if overpredicted else "negative"
    else:
        kwargs["direction"] = "negative" if overpredicted else "positive"

    # Set expectations to mean targets for non-nominal modes
    if expectations is None and mode != "nominal":
        expectations = pd.Series(observations.mean(), index=observations.index)

    # Set appropriate scoring function
    if scoring == "Bernoulli":
        scoring = Bernoulli(**kwargs)
    elif scoring == "BerkJones":
        scoring = BerkJones(**kwargs)
    elif scoring == "Gaussian":
        scoring = Gaussian(**kwargs)
    elif scoring == "Poisson":
        scoring = Poisson(**kwargs)
    else:
        scoring = scoring(**kwargs)

    if mode == "binary": # Flip observations if favorable_value is 0 in binary mode.
        observations = pd.Series(observations == favorable_value, dtype=int)
    elif mode == "nominal":
        unique_outs = set(sorted(observations.unique()))
        size_unique_outs = len(unique_outs)
        if expectations is not None: # Set expectations to 1/(num of categories) for nominal mode
            expectations_cols = set(sorted(expectations.columns))
            assert (
                unique_outs == expectations_cols
            ), f"Expected {unique_outs} in expectation columns, got {expectations_cols}"
        else:
            expectations = pd.Series(
                1 / observations.nunique(), index=observations.index
            )
        max_nominal = kwargs.get("max_nominal", 10)

        assert (
            size_unique_outs <= max_nominal
        ), f"Nominal mode only support up to {max_nominal} labels, got {size_unique_outs}. Use keyword argument max_nominal to increase the limit."

        if favorable_value != "flag-all": # If favorable flag is set, use one-vs-others strategy to scan, else use one-vs-all strategy
            observations = observations.map({favorable_value: 1})
            observations = observations.fillna(0)
            if isinstance(expectations, pd.DataFrame):
                expectations = expectations[favorable_value]
        else:
            results = {}
            orig_observations = observations.copy()
            orig_expectations = expectations.copy()
            for unique in uniques:
                observations = orig_observations.map({unique: 1})
                observations = observations.fillna(0)

                if isinstance(expectations, pd.DataFrame):
                    expectations = orig_expectations[unique]

                scanner = MDSS(scoring)
                result = scanner.scan(
                    data, expectations, observations, penalty, num_iters, mode=mode
                )
                results[unique] = result
            return results

    scanner = MDSS(scoring)
    return scanner.scan(data, expectations, observations, penalty, num_iters, mode=mode)



import os
import sys

import numpy as np
import pytest

from aif360.algorithms.preprocessing import LFR
from aif360.algorithms.preprocessing.optim_preproc_helpers.data_preproc_functions import load_preproc_data_adult


sys.path.append("../../../")
sys.path.append(os.path.dirname(__file__))


@pytest.fixture(scope="module")
def lfrAlgoInstance():
    """This fixture creates two functions with the scope module lfrAlgoInstance creates an instance of the LFR that can
     used by both fit and transform functions.
    ad creates a adult data set that will be used by the fit and the transform functions.
    """
    privileged_groups = [{'sex': 1.0}]
    unprivileged_groups = [{'sex': 0.0}]

    lfrAlgoInstance = LFR(unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)

    return lfrAlgoInstance


@pytest.fixture(scope="module")
def ad():
    return load_preproc_data_adult().split([0.7], shuffle=True)[0]


@pytest.fixture(scope="module")
def lfrfitmodel():
    privileged_groups = [{'sex': 1.0}]
    unprivileged_groups = [{'sex': 0.0}]

    lfrAlgoInstance = LFR(unprivileged_groups=unprivileged_groups, privileged_groups=privileged_groups)
    ad = load_preproc_data_adult().split([0.7], shuffle=True)[0]
    TR = lfrAlgoInstance.fit(ad)

    return TR


def test_fit_isnumpy(lfrfitmodel):
    """The Fit function returns a numpy and it should asserted whether it really returned a numpy precision 64 bits.
    """
    expected = True
    if type(lfrfitmodel.learned_model) is np.ndarray:
        res = True
    else:
        res = False
    assert res == expected


def test_fit_notnull(lfrfitmodel):
    """Should not be null.
    """
    expected = False
    if lfrfitmodel.learned_model is None:
        res = True
    else:
        res = False
    print("numpy:" + str(res))
    assert res == expected


def test_fit_notallzeros(lfrfitmodel):
    """Should not be all zeros.
    """
    expected = False
    all_zeros = not np.any(lfrfitmodel)
    print("allzeros:" + str(all_zeros))
    assert all_zeros == expected


def test_fit_notNaN(lfrfitmodel):
    """Should not have nan's in it.
    """
    expected = False
    res = np.isnan(lfrfitmodel.learned_model).any()
    print("nan:" + str(res))
    assert res == expected


# --------------------------------------------------------------------#
# Transform function testing methods
# --------------------------------------------------------------------#

def test_transform_protecteddataset(lfrfitmodel, ad):
    """After transformation - it should not change protected attributes - it should be same as input.
    """
    lftransformeddataset = lfrfitmodel.transform(ad, threshold=0.3)
    # print( ad.protected_attributes)
    print(type(lftransformeddataset.protected_attributes))
    # print("transformeddataset:" + lfttransformeddataset.protected_attributes)
    expected = True
    res = np.array_equal(lftransformeddataset.protected_attributes, ad.protected_attributes)
    assert res == expected


def test_transform_notNaN(lfrfitmodel, ad):
    """The transformed data should not have any columns or rows summing upto zero.
    """
    lftransformeddataset = lfrfitmodel.transform(ad, threshold=0.3)
    lstrowsum = np.sum(lftransformeddataset.features, axis=1).tolist()
    expected = False
    allrow_zeros = not np.any(lstrowsum)
    lstcolsum = np.sum(lftransformeddataset.features, axis=0).tolist()
    allcol_zeros = not np.any(lstcolsum)
    assert (allrow_zeros and allcol_zeros) == expected


def test_transform_notNaN2(lfrfitmodel, ad):
    """The transformed data should not contain nan's. Using the threshold value of 0.3.
    """
    lftransformeddataset = lfrfitmodel.transform(ad, threshold=0.3)
    expected = False
    res = np.isnan(lftransformeddataset.features).any()
    assert res == expected



from aif360.detectors.mdss.ScoringFunctions.ScoringFunction import ScoringFunction
from aif360.detectors.mdss.ScoringFunctions import optim

import numpy as np


class Poisson(ScoringFunction):
    def __init__(self, **kwargs):
        """
        Poisson score function. May be appropriate to use when the outcome of
        interest is assumed to be Poisson distributed or Binary.

        kwargs must contain
        'direction (str)' - direction of the severity; could be higher than expected outcomes ('positive') or lower than expected ('negative')
        """

        super(Poisson, self).__init__(**kwargs)

    def score(self, observed_sum: float, expectations: np.array, penalty: float, q: float):
        """
        Computes poisson bias score for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty term. Should be positive
        :param q: current value of q
        :return: bias score for the current value of q
        """

        assert q > 0, (
            "Warning: calling compute_score_given_q with "
            "observed_sum=%.2f, expectations of length=%d, penalty=%.2f, q=%.2f"
            % (observed_sum, len(expectations), penalty, q)
        )

        ans = observed_sum * np.log(q) + (expectations - q * expectations).sum() - penalty
        return ans

    def qmle(self, observed_sum: float, expectations: np.array):
        """
        Computes the q which maximizes score (q_mle).
        """
        direction = self.direction
        ans = optim.bisection_q_mle(self, observed_sum, expectations, direction=direction)
        return ans

    def compute_qs(self, observed_sum: float, expectations: np.array, penalty: float):
        """
        Computes roots (qmin and qmax) of the score function for given q

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param penalty: penalty coefficient
        """

        direction = self.direction
        q_mle = self.qmle(observed_sum, expectations)

        if self.score(observed_sum, expectations, penalty, q_mle) > 0:
            exist = 1
            q_min = optim.bisection_q_min(self, observed_sum, expectations, penalty, q_mle)
            q_max = optim.bisection_q_max(self, observed_sum, expectations, penalty, q_mle)
        else:
            # there are no roots
            exist = 0
            q_min = 0
            q_max = 0

        # only consider the desired direction, positive or negative
        if exist:
            exist, q_min, q_max = optim.direction_assertions(direction, q_min, q_max)

        ans = [exist, q_mle, q_min, q_max]
        return ans

    def q_dscore(self, observed_sum, expectations, q):
        """
        This actually computes q times the slope, which has the same sign as the slope since q is positive.
        score = Y log q + \sum_i (p_i - qp_i)
        dscore/dq = Y / q - \sum_i(p_i)
        q dscore/dq = q_dscore = Y - (q * \sum_i(p_i))

        :param observed_sum: sum of observed binary outcomes for all i
        :param expectations: predicted outcomes for each data element i
        :param q: current value of q
        :return: q dscore/dq
        """
        ans = observed_sum - (q * expectations).sum()
        return ans



from functools import partial

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_almost_equal
from sklearn.linear_model import LogisticRegression

from aif360.datasets import AdultDataset
from aif360.sklearn.datasets import fetch_adult
from aif360.metrics import ClassificationMetric
from aif360.sklearn.metrics import (
        consistency_score, specificity_score, selection_rate,
        base_rate, smoothed_base_rate, generalized_fpr, generalized_fnr,
        disparate_impact_ratio, statistical_parity_difference,
        equal_opportunity_difference, average_odds_difference, average_predictive_value_difference,
        average_odds_error, smoothed_edf, df_bias_amplification,
        generalized_entropy_error, between_group_generalized_entropy_error,
        class_imbalance, kl_divergence, conditional_demographic_disparity,
        intersection, one_vs_rest, make_scorer)


X, y, sample_weight = fetch_adult(numeric_only=True)
lr = LogisticRegression(solver='liblinear').fit(X, y, sample_weight=sample_weight)
y_pred = lr.predict(X)
y_proba = lr.predict_proba(X)[:, 1]
adult = AdultDataset(instance_weights_name='fnlwgt', categorical_features=[],
        features_to_keep=['age', 'education-num', 'capital-gain',
                          'capital-loss', 'hours-per-week'],
        features_to_drop=[])
adult_pred = adult.copy()
adult_pred.labels = y_pred
adult_pred.scores = y_proba
cm = ClassificationMetric(adult, adult_pred,
                          unprivileged_groups=[{'sex': 0}],
                          privileged_groups=[{'sex': 1}])

def test_dataset_equality():
    """Tests that the old and new datasets match exactly."""
    assert (adult.features == X.values).all()
    assert (adult.labels.ravel() == y).all()

def test_consistency():
    """Tests that the old and new consistency_score matches exactly."""
    assert np.isclose(consistency_score(X, y), cm.consistency())

def test_specificity():
    """Tests that the old and new specificity_score matches exactly."""
    spec = specificity_score(y, y_pred, sample_weight=sample_weight)
    assert spec == cm.specificity()

def test_base_rate():
    """Tests that the old and new base_rate matches exactly."""
    base = base_rate(y, y_pred, sample_weight=sample_weight)
    assert base == cm.base_rate()

def test_selection_rate():
    """Tests that the old and new selection_rate matches exactly."""
    select = selection_rate(y, y_pred, sample_weight=sample_weight)
    assert select == cm.selection_rate()

def test_smoothed_base_rate():
    sbr = smoothed_base_rate(y, y_pred, concentration=0, sample_weight=sample_weight)
    assert sbr == base_rate(y, y_pred, sample_weight=sample_weight)

    sbr_int = intersection(smoothed_base_rate, y, y_pred, sample_weight=sample_weight)
    assert (sbr_int == cm._smoothed_base_rates(cm.dataset.labels)).all()

def test_generalized_fpr():
    """Tests that the old and new generalized_fpr matches exactly."""
    gfpr = generalized_fpr(y, y_proba, sample_weight=sample_weight)
    assert np.isclose(gfpr, cm.generalized_false_positive_rate())

def test_generalized_fnr():
    """Tests that the old and new generalized_fnr matches exactly."""
    gfnr = generalized_fnr(y, y_proba, sample_weight=sample_weight)
    assert np.isclose(gfnr, cm.generalized_false_negative_rate())

def test_disparate_impact():
    """Tests that the old and new disparate_impact matches exactly."""
    di = disparate_impact_ratio(y, y_pred, prot_attr='sex',
                                sample_weight=sample_weight)
    assert di == cm.disparate_impact()

def test_statistical_parity():
    """Tests that the old and new statistical_parity matches exactly."""
    stat = statistical_parity_difference(y, y_pred, prot_attr='sex',
                                         sample_weight=sample_weight)
    assert stat == cm.statistical_parity_difference()

def test_equal_opportunity():
    """Tests that the old and new equal_opportunity matches exactly."""
    eopp = equal_opportunity_difference(y, y_pred, prot_attr='sex',
                                        sample_weight=sample_weight)
    assert eopp == cm.equal_opportunity_difference()

def test_average_odds_difference():
    """Tests that the old and new average_odds_difference matches exactly."""
    aod = average_odds_difference(y, y_pred, prot_attr='sex',
                                  sample_weight=sample_weight)
    assert np.isclose(aod, cm.average_odds_difference())

def test_average_predictive_value_difference():
    """Tests that the old and new average_predictive_value_difference matches exactly."""
    aod = average_predictive_value_difference(y, y_pred, prot_attr='sex',
                                  sample_weight=sample_weight)
    assert np.isclose(aod, cm.average_predictive_value_difference())

def test_average_odds_error():
    """Tests that the old and new average_odds_error matches exactly."""
    aoe = average_odds_error(y, y_pred, prot_attr='sex',
                             sample_weight=sample_weight)
    assert np.isclose(aoe, cm.average_abs_odds_difference())

def test_smoothed_edf():
    """Tests that the old and new smoothed_edf matches exactly."""
    edf = smoothed_edf(y, sample_weight=sample_weight)
    assert edf == cm.smoothed_empirical_differential_fairness()

    edf = smoothed_edf(y, concentration=1e9, sample_weight=sample_weight)
    assert edf == cm.smoothed_empirical_differential_fairness(1e9)

def test_df_bias_amplification():
    """Tests that the old and new df_bias_amplification matches exactly."""
    amp = df_bias_amplification(y, y_pred, sample_weight=sample_weight)
    assert amp == cm.differential_fairness_bias_amplification()

def test_generalized_entropy_index():
    """Tests that the old and new generalized_entropy_index matches exactly."""
    gei = generalized_entropy_error(y, y_pred)
    assert np.isclose(gei, cm.generalized_entropy_index())

def test_between_group_generalized_entropy_index():
    """Tests that the old and new between_group_GEI matches exactly."""
    bggei = between_group_generalized_entropy_error(y, y_pred, prot_attr='sex')
    assert bggei == cm.between_group_generalized_entropy_index()

def test_class_imbalance():
    prot_attr = pd.Series([1, 1, 1, 1, 0, 0, 0], name='sex')
    y = pd.Series(np.random.random(7), index=prot_attr)  # y values are irrelevant
    sample_weight = np.array([1, 2, 3, 4, 3, 2, 1])
    assert class_imbalance(y, sample_weight=sample_weight) == -0.25  # -4/16
    assert class_imbalance(y, priv_group=0, sample_weight=sample_weight) == 0.25

def test_kl_divergence():
    prot_attr = pd.Series([1, 1, 1, 1, 0, 0, 0], name='sex')
    y = pd.Series([0, 1, 2, 0, 0, 1, 2], index=prot_attr)
    sample_weight = [1, 2, 3, 4, 4, 3, 3]
    kld = kl_divergence(y, priv_group=1, sample_weight=sample_weight)
    assert np.isclose(kld, (5*np.log(5/4) + 2*np.log(2/3) + 3*np.log(3/3))/10)

    kld = kl_divergence(y, priv_group=0, sample_weight=sample_weight)
    assert np.isclose(kld, (4*np.log(4/5) + 3*np.log(3/2) + 3*np.log(3/3))/10)

def test_conditional_demographic_disparity():
    prot_attr = pd.Series([0, 0, 1, 1, 2, 2, 2], name='sex')
    y = pd.Series([0, 1, 0, 1, 0, 1, 0], index=prot_attr)
    sample_weight = [1, 2, 3, 4, 3, 2, 1]
    cdd = conditional_demographic_disparity(y, sample_weight=sample_weight)
    assert cdd == (3*(1-2) + 7*(3-4) + 6*(4-2)) / (8*16)

@pytest.mark.parametrize(
    "func, is_ratio",
    [
        (statistical_parity_difference, False),
        (disparate_impact_ratio, True),
        (equal_opportunity_difference, False),
        (average_odds_difference, False),
    ],
)
def test_make_scorer(func, is_ratio):
    actual = func(y, y_pred, prot_attr="sex", priv_group=1)
    actual_fliped = func(y, y_pred, prot_attr="sex", priv_group=0)
    scorer = make_scorer(func, is_ratio=is_ratio, prot_attr="sex", priv_group=1)
    expected = scorer(lr, X, y)
    if is_ratio:
        ret = min(actual, actual_fliped)
        assert_almost_equal(ret, expected, 3)
    else:
        # The lower the better
        assert_almost_equal(-abs(actual), expected, 3)
        assert_almost_equal(-abs(actual_fliped), expected, 3)

@pytest.mark.parametrize(
    "func",
    [
        disparate_impact_ratio,  # ratio
        average_odds_error,  # difference
        between_group_generalized_entropy_error,  # index
    ]
)
def test_explicit_prot_attr_array(func):
    """Tests that metrics work with explicit prot_attr arrays."""
    prot_attr = y.index.to_flat_index()#y.index.get_level_values('sex')
    y_arr = y.to_numpy()
    f = partial(func, priv_group=(1, 1))
    assert f(y_arr, y_pred, prot_attr=prot_attr) == f(y, y_pred)

def test_one_vs_rest():
    ovr = one_vs_rest(statistical_parity_difference, y, y_pred, prot_attr='sex')
    assert ovr[0] == -ovr[1]
    assert ovr[1] == statistical_parity_difference(y, y_pred, prot_attr='sex')

    v, k = one_vs_rest(disparate_impact_ratio, y, y_pred, return_groups=True)
    assert len(v) == 4
    ovr = dict(zip(k, v))
    for i in range(2):
        for j in range(2):
            g = (i, j)
            assert disparate_impact_ratio(y, y_pred, priv_group=g) == ovr[g]



import pandas as pd
import numpy as np


def get_entire_subset():
    """
    Returns the entire subset, which is an empty dictionary
    :return: empty dictionary
    """
    return {}


def get_random_subset(coordinates: pd.DataFrame, prob: float, min_elements: int = 0):
    """
    Returns a random subset
    :param coordinates: data frame containing having as columns the features
    :param prob: probability to select a value of a feature
    :param min_elements: minimum number of elements to be included in the randomly generated sub-population
    :return: dictionary representing a random sub-population
    """

    subset_random_values = {}
    shuffled_column_names = np.random.permutation(coordinates.columns.values)

    # consider each column once, in random order
    for column_name in shuffled_column_names:
        # get unique values of the current column
        temp = coordinates[column_name].unique()

        # include each attribute value with probability = prob
        mask_values = np.random.rand(len(temp)) < prob

        if mask_values.sum() < len(temp):
            # set values for the current column
            subset_random_values[column_name] = temp[mask_values].tolist()

            # compute the remaining records
            mask_subset = coordinates[subset_random_values.keys()].isin(subset_random_values).all(axis=1)
            remaining_records = len(coordinates.loc[mask_subset])

            # only filter on this attribute if at least min_elements records would be kept
            if remaining_records < min_elements:
                del subset_random_values[column_name]

    return subset_random_values


from abc import ABCMeta


def do_decorate(attr, value):
    return not attr.startswith('_') and callable(value) and getattr(value, '__decorate__', True)

def factory(decorator):
    class ApplyDecoratorMeta(ABCMeta):
        """Metaclass which applies a decorator to all public, non-special
        instance methods.

        Note:
            `decorator` must use @functools.wraps(f) for abstractmethod to work.

        https://stackoverflow.com/questions/10067262/automatically-decorating-every-instance-method-in-a-class
        """
        def __new__(cls, name, bases, dct):
            for attr, value in dct.items():
                if do_decorate(attr, value):
                    dct[attr] = decorator(value)
            return super(ApplyDecoratorMeta, cls).__new__(cls, name, bases, dct)
    return ApplyDecoratorMeta

def dont_decorate(func):
    func.__decorate__ = False
    return func

def ApplyDecorator(decorator):
    return factory(decorator)(str('ApplyDecorator'), (), {})



import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline

from aif360.algorithms.preprocessing import LFR
from aif360.sklearn.preprocessing import LearnedFairRepresentations
from aif360.sklearn.metrics import make_scorer, statistical_parity_difference


@pytest.fixture
def old_lfr(old_german):
    lfr = LFR(unprivileged_groups=[{'age': 0}], privileged_groups=[{'age': 1}],
              seed=123)
    return lfr.fit(old_german, maxfun=3e4)

@pytest.fixture
def old_lfr2(old_german):
    lfr = LFR(unprivileged_groups=[{'age': 0}], privileged_groups=[{'age': 1}],
              seed=123)
    return lfr.fit_transform(old_german, maxfun=3e4)

@pytest.fixture
def new_lfr(new_german):
    lfr = LearnedFairRepresentations('age', random_state=123)
    return lfr.fit(**new_german._asdict())

@pytest.fixture
def new_lfr2(new_german):
    lfr = LearnedFairRepresentations('age', random_state=123)
    X, y = new_german
    Xt = lfr.fit_transform(X, y)
    yt = lfr.predict(X)
    return Xt, yt


def test_data_old_new(old_german, new_german):
    """Test that the old and new versions of the dataset match."""
    assert (old_german.features == new_german.X).all(None)
    assert (old_german.labels.flatten() == new_german.y).all()

def test_lfr_old_reproduce(old_lfr, old_lfr2, old_german):
    """Test that the old LFR is reproducible."""
    old_lfr = old_lfr.transform(old_german)
    assert np.allclose(old_lfr.features, old_lfr2.features)
    assert np.allclose(old_lfr.labels, old_lfr2.labels)

def test_lfr_new_reproduce(new_lfr, new_lfr2, new_german):
    """Test that the new LearnedFairRepresentation is reproducible."""
    Xt = new_lfr.transform(new_german.X)
    yt = new_lfr.predict(new_german.X)
    Xt2, yt2 = new_lfr2

    assert np.allclose(Xt, Xt2)
    assert np.allclose(yt, yt2)

@pytest.mark.skip()
def test_lfr_models(old_lfr, new_lfr):
    """Test that the learned model parameters of the old and new LFR match."""
    assert np.allclose(old_lfr.w, new_lfr.coef_.flatten())
    assert np.allclose(old_lfr.prototypes, new_lfr.prototypes_)

@pytest.mark.skip()
def test_lfr_old_new(old_lfr2, new_lfr2):
    """Test that the transformations of the old and new LFR match."""
    Xt, yt = new_lfr2
    assert np.allclose(old_lfr2.features, Xt)
    assert np.allclose(old_lfr2.labels.flatten(), yt)

def test_lfr_multiclass():
    """Test that the new LFR runs with >2 classes."""
    X, y = make_classification(n_informative=3, n_classes=4)
    prot_attr = np.random.randint(2, size=X.shape[:1])
    X = pd.DataFrame(X, index=pd.Series(prot_attr, name='prot_attr'))
    lfr = LearnedFairRepresentations()
    lfr.fit(X, y)
    assert lfr.coef_.shape[1] == 4

def test_lfr_grid(new_german):
    """Test that the new LFR works in a grid search (and that debiasing
    results in improved statistical parity difference).
    """
    X, y = new_german
    lfr = LearnedFairRepresentations('age', reconstruct_weight=0, max_iter=1000)
    params = {'fairness_weight': [50, 0]}
    min_disc = make_scorer(statistical_parity_difference, prot_attr='age')
    clf = GridSearchCV(lfr, params, scoring=min_disc, cv=2)
    clf.fit(X, y)

    assert clf.best_params_ == {'fairness_weight': 50}

def test_lfr_pipe(new_german):
    """Test that the new LFR works as a pre-processing step in a pipeline."""
    X, y = new_german
    lfr = LearnedFairRepresentations('age', random_state=123)
    lr = LogisticRegression(solver='lbfgs', random_state=123)
    pipe = make_pipeline(lfr, lr)

    pipe.fit(X, y)



from itertools import permutations
from typing import Union
import warnings

import numpy as np
import pandas as pd
from scipy.special import rel_entr
from sklearn.metrics import make_scorer as _make_scorer, recall_score, precision_score
from sklearn.metrics import multilabel_confusion_matrix
from sklearn.metrics._classification import _prf_divide, _check_zero_division
from sklearn.neighbors import NearestNeighbors
from sklearn.utils import check_X_y

from aif360.metrics import ot_metric
from aif360.sklearn.utils import check_inputs, check_groups
from aif360.detectors.mdss.ScoringFunctions import BerkJones, Bernoulli
from aif360.detectors.mdss.MDSS import MDSS


# ============================= META-METRICS ===================================
def difference(func, y_true, y_pred=None, prot_attr=None, priv_group=1,
               sample_weight=None, **kwargs):
    """Compute the difference between unprivileged and privileged subsets for an
    arbitrary metric.

    Note: The optimal value of a difference is 0. To make it a scorer, one must
    take the absolute value and set greater_is_better to False.

    Unprivileged group is taken to be the inverse of the privileged group.

    Args:
        func (function): A metric function from :mod:`sklearn.metrics` or
            :mod:`aif360.sklearn.metrics`.
        y_true (pandas.Series): Outcome vector with protected attributes as
            index.
        y_pred (array-like, optional): Estimated outcomes.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y are used.
        priv_group (scalar, optional): The label of the privileged group.
        sample_weight (array-like, optional): Sample weights passed through to
            func.
        **kwargs: Additional keyword args to be passed through to func.

    Returns:
        scalar: Difference in metric value for unprivileged and privileged
        groups.

    Examples:
        >>> X, y = fetch_german(numeric_only=True)
        >>> y_pred = LogisticRegression().fit(X, y).predict(X)
        >>> difference(precision_score, y, y_pred, prot_attr='sex',
        ... priv_group='male')
        -0.06955430006277463
    """
    groups, _ = check_groups(y_true, prot_attr)
    idx = (groups == priv_group)
    unpriv = [y[~idx] for y in (y_true, y_pred) if y is not None]
    priv = [y[idx] for y in (y_true, y_pred) if y is not None]
    if sample_weight is not None:
        sample_weight = np.asarray(sample_weight)
        return (func(*unpriv, sample_weight=sample_weight[~idx], **kwargs)
              - func(*priv, sample_weight=sample_weight[idx], **kwargs))
    return func(*unpriv, **kwargs) - func(*priv, **kwargs)

def ratio(func, y_true, y_pred=None, prot_attr=None, priv_group=1,
          sample_weight=None, zero_division='warn', **kwargs):
    """Compute the ratio between unprivileged and privileged subsets for an
    arbitrary metric.

    Note: The optimal value of a ratio is 1. To make it a scorer, one must
    take the minimum of the ratio and its inverse.

    Unprivileged group is taken to be the inverse of the privileged group.

    Args:
        func (function): A metric function from :mod:`sklearn.metrics` or
            :mod:`aif360.sklearn.metrics`.
        y_true (pandas.Series): Outcome vector with protected attributes as
            index.
        y_pred (array-like, optional): Estimated outcomes.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y are used.
        priv_group (scalar, optional): The label of the privileged group.
        sample_weight (array-like, optional): Sample weights passed through to
            func.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.
        **kwargs: Additional keyword args to be passed through to func.

    Returns:
        scalar: Ratio of metric values for unprivileged and privileged groups.
    """
    _check_zero_division(zero_division)
    groups, _ = check_groups(y_true, prot_attr)
    idx = (groups == priv_group)
    unpriv = [y[~idx] for y in (y_true, y_pred) if y is not None]
    priv = [y[idx] for y in (y_true, y_pred) if y is not None]
    if sample_weight is not None:
        sample_weight = np.asarray(sample_weight)
        numerator = func(*unpriv, sample_weight=sample_weight[~idx], **kwargs)
        denominator = func(*priv, sample_weight=sample_weight[idx], **kwargs)
    else:
        numerator = func(*unpriv, **kwargs)
        denominator = func(*priv, **kwargs)

    if func == base_rate:
        modifier = 'positive privileged'
    elif func == selection_rate:
        modifier = 'predicted privileged'
    else:
        modifier = f'value for {func.__name__} on privileged'
    return _prf_divide(np.array([numerator]), np.array([denominator]), 'ratio',
                       modifier, None, ('ratio',), zero_division).item()

def intersection(func, y_true, y_pred=None, prot_attr=None, sample_weight=None,
                 return_groups=False, **kwargs):
    """Compute an arbitrary metric on all intersectional groups of the protected
    attributes provided.

    Args:
        func (function): A metric function from :mod:`sklearn.metrics` or
            :mod:`aif360.sklearn.metrics`.
        y_true (pandas.Series): Outcome vector with protected attributes as
            index.
        y_pred (array-like, optional): Estimated outcomes.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y are used.
        sample_weight (array-like, optional): Sample weights passed through to
            func.
        return_groups (bool, optional): Return group names in addition to metric
            values. Names are tuples of protected attribute values.
        **kwargs: Additional keyword args to be passed through to func.

    Returns:
        list: List of metric values for each intersectional group.

        tuple:
            Metric values and their corresponding group names.

            * **vals** (`list`) -- List of metric values for each intersectional
              group
            * **groups** (:class:`numpy.ndarray`) -- Array of tuples containing
              unique intersectional groups derived from the provided protected
              attributes.

    Examples:
        >>> X, y = fetch_german()
        >>> v, k = intersection(base_rate, y, prot_attr=['sex', 'age'],
        ...                     return_groups=True, pos_label='good')
        >>> dict(zip(k, v))
        {('female', 'aged'): 0.697560975609756,
         ('female', 'young'): 0.5523809523809524,
         ('male', 'aged'): 0.7388429752066116,
         ('male', 'young'): 0.611764705882353}
    """
    groups, _ = check_groups(y_true, prot_attr)
    unique_groups = np.unique(groups)
    func_vals = []
    for g in unique_groups:
        idx = (groups == g)
        sub = [y[idx] for y in (y_true, y_pred) if y is not None]
        if sample_weight is not None:
            sample_weight = np.asarray(sample_weight)
            func_vals.append(func(*sub, sample_weight=sample_weight[idx],
                                  **kwargs))
        else:
            func_vals.append(func(*sub, **kwargs))
    if return_groups:
        return func_vals, unique_groups
    return func_vals

def one_vs_rest(func, y_true, y_pred=None, prot_attr=None, return_groups=False,
                **kwargs):
    """Compute an arbitrary difference/ratio metric on all intersectional groups
    of the protected attributes provided in a one-vs-rest manner.

    Args:
        func (function): A difference or ratio metric function from
            :mod:`aif360.sklearn.metrics`.
        y_true (pandas.Series): Outcome vector with protected attributes as
            index.
        y_pred (array-like, optional): Estimated outcomes.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y are used.
        sample_weight (array-like, optional): Sample weights passed through to
            func.
        return_groups (bool): Return group names in addition to metric values.
            Names are tuples of protected attribute values.
        **kwargs: Additional keyword args to be passed through to func.

    Returns:
        list: List of metric values considering each intersectional group in
        turn as privileged and the rest as unprivileged.

        tuple:
            Metric values and their corresponding group names.

            * **vals** (`list`) -- List of metric values considering each
              group in turn as privileged and the rest as unprivileged.
            * **groups** (:class:`numpy.ndarray`) -- Array of tuples containing
              unique intersectional groups derived from the provided protected
              attributes.

    Examples:
        >>> X, y = fetch_german()
        >>> v, k = one_vs_rest(statistical_parity_difference, y,
        ...                    prot_attr=['sex', 'age'], return_groups=True,
        ...                    pos_label='good')
        >>> dict(zip(k, v))
        {(0, 0): 0.16493748337323755,
         (0, 1): 0.0030679552078539674,
         (1, 0): 0.09643201542912239,
         (1, 1): -0.09833664609268755}

        >>> from functools import partial
        >>> from sklearn.metrics import accuracy_score
        >>> from sklearn.linear_model import LogisticRegression
        >>> y_pred = LogisticRegression(solver='liblinear').fit(X, y).predict(X)
        >>> acc_diff = partial(difference, accuracy_score)
        >>> one_vs_rest(acc_diff, y, y_pred, prot_attr=['sex', 'age'])
        [0.11338121840915127,
         -0.013775118883264326,
         0.018450658952105403,
         -0.04119677790563869]
    """
    groups, _ = check_groups(y_true, prot_attr)
    unique_groups = np.unique(groups)
    func_vals = []
    for g in unique_groups:
        func_vals.append(func(y_true, y_pred, prot_attr=prot_attr, priv_group=g,
                              **kwargs))
    if return_groups:
        return func_vals, unique_groups
    return func_vals


# =========================== SCORER FACTORY =================================
def make_scorer(score_func, is_ratio=False, **kwargs):
    """Make a scorer from a 'difference' or 'ratio' metric (e.g.
    :func:`statistical_parity_difference`).

    Args:
        score_func (callable): A ratio/difference metric with signature
            ``score_func(y, y_pred, **kwargs)``.
        is_ratio (boolean, optional): Indicates if the metric is ratio or
        difference based.
    """
    if is_ratio:

        def score(y, y_pred, **kwargs):
            ratio = score_func(y, y_pred, **kwargs)
            eps = np.finfo(float).eps
            ratio_inverse = 1 / ratio if ratio > eps else eps
            return min(ratio, ratio_inverse)

        scorer = _make_scorer(score, **kwargs)
    else:

        def score(y, y_pred, **kwargs):
            diff = score_func(y, y_pred, **kwargs)
            return abs(diff)

        scorer = _make_scorer(score, greater_is_better=False, **kwargs)
    return scorer

# ================================ HELPERS =====================================
def num_samples(y_true, y_pred=None, sample_weight=None):
    """Compute the number of samples.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets. Ignored.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: (Weighted) number of samples.
    """
    sample_weight = check_inputs(y_true, y_true, sample_weight, ensure_2d=False)[2]
    return sum(sample_weight)

def num_pos_neg(y_true, y_pred=None, pos_label=1, sample_weight=None):
    """Compute the number of positive and negative samples.

    Args:
        y_true (array-like): Ground truth (correct) target values. If y_pred is
            provided, this is ignored.
        y_pred (array-like): Estimated targets as returned by a classifier.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        tuple:
            Number of positives and negatives.

            * **n_positive** (`float`) -- (Weighted) number of positive samples.
            * **n_negative** (`float`) -- (Weighted) number of negative samples.
    """
    y = y_true if y_pred is None else y_pred
    sample_weight = check_inputs(y_true, y, sample_weight, ensure_2d=False)[2]
    pos = (y == pos_label).tolist()
    neg = (y != pos_label).tolist()
    return sum(sample_weight[pos]), sum(sample_weight[neg])

def specificity_score(y_true, y_pred, *, pos_label=1, sample_weight=None,
                      zero_division='warn'):
    """Compute the specificity or true negative rate.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.
    """
    _check_zero_division(zero_division)
    MCM = multilabel_confusion_matrix(y_true, y_pred, labels=[pos_label],
                                      sample_weight=sample_weight)
    tn, fp = MCM[:, 0, 0], MCM[:, 0, 1]
    negs = tn + fp
    return _prf_divide(tn, negs, 'specificity', 'negative', None,
                       ('specificity',), zero_division).item()

def false_omission_rate_error(y_true, y_pred, *, pos_label=1, sample_weight=None,
                      zero_division='warn'):
    """Compute the false omission rate.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.
    """
    _check_zero_division(zero_division)
    MCM = multilabel_confusion_matrix(y_true, y_pred, labels=[pos_label],
                                      sample_weight=sample_weight)
    tn, fn = MCM[:, 0, 0], MCM[:, 1, 0]
    negs = tn + fn
    return _prf_divide(fn, negs, 'false omission rate', 'predicted negative', None,
                       ('false omission rate',), zero_division).item()

def base_rate(y_true, y_pred=None, *, pos_label=1, sample_weight=None):
    r"""Compute the base rate, :math:`Pr(Y = \text{pos_label}) = \frac{P}{P+N}`.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like, optional): Estimated targets. Ignored.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Base rate.
    """
    idx = (y_true == pos_label)
    return np.average(idx, weights=sample_weight)

def selection_rate(y_true, y_pred, *, pos_label=1, sample_weight=None):
    r"""Compute the selection rate, :math:`Pr(\hat{Y} = \text{pos_label}) =
    \frac{TP + FP}{P + N}`.

    Args:
        y_true (array-like): Ground truth (correct) target values. Ignored.
        y_pred (array-like): Estimated targets as returned by a classifier.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Selection rate.
    """
    return base_rate(y_pred, pos_label=pos_label, sample_weight=sample_weight)

def smoothed_base_rate(y_true, y_pred=None, *, concentration=1.0, pos_label=1,
                       sample_weight=None):
    r"""Compute the smoothed base rate,
    :math:`\frac{P + \alpha}{P + N + |R_Y|\alpha}`.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like, optional): Estimated targets. Ignored.
        concentration (scalar): Dirichlet smoothing concentration parameter
            :math:`|R_Y|\alpha` (must be non-negative).
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Smoothed base rate.
    """
    if concentration < 0:
        raise ValueError("Concentration parameter must be non-negative.")
    num_classes = len(np.unique(y_true))
    idx = (y_true == pos_label)
    avg, tot = np.average(idx, weights=sample_weight, returned=True)
    return (avg*tot + concentration/num_classes) / (tot + concentration)

def smoothed_selection_rate(y_true, y_pred, *, concentration=1.0, pos_label=1,
                            sample_weight=None):
    r"""Compute the smoothed selection rate,
    :math:`\frac{TP + FP + \alpha}{P + N + |R_Y|\alpha}`.

    Args:
        y_true (array-like): Ground truth (correct) target values. Ignored.
        y_pred (array-like): Estimated targets as returned by a classifier.
        concentration (scalar): Dirichlet smoothing concentration parameter
            :math:`|R_Y|\alpha` (must be non-negative).
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Smoothed selection rate.
    """
    return smoothed_base_rate(y_pred, concentration=concentration,
                              pos_label=pos_label, sample_weight=sample_weight)

def generalized_fpr(y_true, probas_pred, *, pos_label=1, sample_weight=None,
                    zero_division='warn'):
    r"""Return the ratio of generalized false positives to negative examples in
    the dataset, :math:`GFPR = \tfrac{GFP}{N}`.

    Generalized confusion matrix measures such as this are calculated by summing
    the probabilities of the positive class instead of the hard predictions.

    Args:
        y_true (array-like): Ground-truth (correct) target values.
        probas_pred (array-like): Probability estimates of the positive class.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.

    Returns:
        float: Generalized false positive rate.
    """
    _check_zero_division(zero_division)
    y_true, probas_pred, sample_weight = check_inputs(y_true, probas_pred,
                                                      sample_weight, False)

    idx = (y_true != pos_label)
    gfp = np.array([np.dot(probas_pred[idx], sample_weight[idx])])
    neg = np.array([sample_weight[idx].sum()])
    return _prf_divide(gfp, neg, 'generalized FPR', 'negative', None,
                       ('generalized FPR',), zero_division).item()

def generalized_fnr(y_true, probas_pred, *, pos_label=1, sample_weight=None,
                    zero_division='warn'):
    r"""Return the ratio of generalized false negatives to positive examples in
    the dataset, :math:`GFNR = \tfrac{GFN}{P}`.

    Generalized confusion matrix measures such as this are calculated by summing
    the probabilities of the positive class instead of the hard predictions.

    Args:
        y_true (array-like): Ground-truth (correct) target values.
        probas_pred (array-like): Probability estimates of the positive class.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.

    Returns:
        float: Generalized false negative rate.
    """
    _check_zero_division(zero_division)
    y_true, probas_pred, sample_weight = check_inputs(y_true, probas_pred,
                                                      sample_weight, False)

    idx = (y_true == pos_label)
    gfn = np.array([np.dot(1 - probas_pred[idx], sample_weight[idx])])
    pos = np.array([sample_weight[idx].sum()])
    return _prf_divide(gfn, pos, 'generalized FNR', 'positive', None,
                       ('generalized FNR',), zero_division).item()


# ============================ GROUP FAIRNESS ==================================
def ot_distance(
    y_true: pd.Series,
    y_pred: Union[pd.Series, pd.DataFrame],
    prot_attr: pd.Series = None,
    pos_label: Union[str, float] = None,
    scoring: str = "Wasserstein1",
    num_iters: int = 1e5,
    penalty: float = 1e-17,
    mode: str = "binary",
    cost_matrix: np.ndarray=None,
    **kwargs,
):
    """Normalize and calculate Wasserstein distance between groups defined by `prot_attr` in `y_true` and `y_pred`.

    Args:
        y_true (pd.Series): ground truth (correct) target values.
        y_pred (pd.Series, pd.DataFrame): estimated target values.
            If `mode` is nominal, must be a `pd.DataFrame` with columns containing predictions for each nominal class,
                or list of corresponding column names in `data`.
            If `None`, model is assumed to be a dummy model that predicts the mean of the targets
                or 1/(number of categories) for nominal mode.
        prot_attr (pd.Series): sensitive attribute values.
            If `None`, assume all samples belong to the same protected group.
        pos_label(str, float, optional): Either "high", "low" or a float value if the mode in [binary, ordinal, or continuous].
                If float, value has to be the minimum or the maximum in the ground_truth column.
                Defaults to high if None for these modes.
                Support for float left in to keep the intuition clear in binary classification tasks.
                If `mode` is nominal, favorable values should be one of the unique categories in the ground_truth.
                Defaults to a one-vs-all scan if None for nominal mode.
        scoring (str or class): only 'Wasserstein1'
        num_iters (int, optional): number of iterations (random restarts) for EMD. Should be positive.
        penalty (float, optional): penalty term. Should be positive. The penalty term as with any regularization parameter
            may need to be tuned for a particular use case. The higher the penalty, the higher the influence of entropy regualizer.
        mode: one of ['binary', 'continuous', 'nominal', 'ordinal']. Defaults to binary.
                In nominal mode, up to 10 categories are supported by default.
                To increase this, pass in keyword argument max_nominal = integer value.
        cost_matrix (np.ndarray): cost matrix for the Wasserstein distance. Defaults to absolute difference between samples.

    Returns:
        ot.emd2 (float, dict): Earth mover's distance or dictionary of optimal transports for each of option of classifier

    Raises:
        ValueError: if `mode` is 'binary' but `ground_truth` contains less than 1 or more than 2 unique values.
    """
    return ot_metric.ot_distance(
        ground_truth=y_true,
        classifier=y_pred,
        prot_attr=prot_attr,
        favorable_value=pos_label,
        scoring=scoring,
        num_iters=num_iters,
        penalty=penalty,
        mode=mode,
        cost_matrix=cost_matrix,
        **kwargs
    )

def statistical_parity_difference(y_true, y_pred=None, *, prot_attr=None,
                                  priv_group=1, pos_label=1, sample_weight=None):
    r"""Difference in selection rates.

    .. math::
        Pr(\hat{Y} = \text{pos_label} | D = \text{unprivileged})
        - Pr(\hat{Y} = \text{pos_label} | D = \text{privileged})

    Note:
        If only y_true is provided, this will return the difference in base
        rates (statistical parity difference of the original dataset). If both
        y_true and y_pred are provided, only y_pred is used.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values. If y_pred
            is provided, this is ignored.
        y_pred (array-like, optional): Estimated targets as returned by a
            classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Statistical parity difference.

    See also:
        :func:`selection_rate`, :func:`base_rate`
    """
    rate = base_rate if y_pred is None else selection_rate
    return difference(rate, y_true, y_pred, prot_attr=prot_attr,
                      priv_group=priv_group, pos_label=pos_label,
                      sample_weight=sample_weight)

def disparate_impact_ratio(y_true, y_pred=None, *, prot_attr=None, priv_group=1,
                           pos_label=1, sample_weight=None, zero_division='warn'):
    r"""Ratio of selection rates.

    .. math::
        \frac{Pr(\hat{Y} = \text{pos_label} | D = \text{unprivileged})}
        {Pr(\hat{Y} = \text{pos_label} | D = \text{privileged})}

    Note:
        If only y_true is provided, this will return the ratio of base rates
        (disparate impact of the original dataset). If both y_true and y_pred
        are provided, only y_pred is used.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values. If y_pred
            is provided, this is ignored.
        y_pred (array-like, optional): Estimated targets as returned by a
            classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.
        zero_division ('warn', 0 or 1): Sets the value to return when there is a
            zero division. If set to “warn”, this acts as 0, but warnings are
            also raised.

    Returns:
        float: Disparate impact.

    See also:
        :func:`selection_rate`, :func:`base_rate`
    """
    rate = base_rate if y_pred is None else selection_rate
    return ratio(rate, y_true, y_pred, prot_attr=prot_attr,
                 priv_group=priv_group, pos_label=pos_label,
                 sample_weight=sample_weight, zero_division=zero_division)

def equal_opportunity_difference(y_true, y_pred, *, prot_attr=None,
                                 priv_group=1, pos_label=1, sample_weight=None):
    r"""A relaxed version of equality of opportunity.

    Returns the difference in recall scores (TPR) between the unprivileged and
    privileged groups. A value of 0 indicates equality of opportunity.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Equal opportunity difference.

    See also:
        :func:`~sklearn.metrics.recall_score`
    """
    return difference(recall_score, y_true, y_pred, prot_attr=prot_attr,
                      priv_group=priv_group, pos_label=pos_label,
                      sample_weight=sample_weight)

def average_odds_difference(y_true, y_pred, *, prot_attr=None, priv_group=1,
                            pos_label=1, sample_weight=None):
    r"""A relaxed version of equality of odds.

    Returns the average of the difference in FPR and TPR for the unprivileged
    and privileged groups:

    .. math::

        \dfrac{(FPR_{D = \text{unprivileged}} - FPR_{D = \text{privileged}})
        + (TPR_{D = \text{unprivileged}} - TPR_{D = \text{privileged}})}{2}

    A value of 0 indicates equality of odds.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Average odds difference.
    """
    fpr_diff = -difference(specificity_score, y_true, y_pred,
                           prot_attr=prot_attr, priv_group=priv_group,
                           pos_label=pos_label, sample_weight=sample_weight)
    tpr_diff = difference(recall_score, y_true, y_pred, prot_attr=prot_attr,
                          priv_group=priv_group, pos_label=pos_label,
                          sample_weight=sample_weight)
    return (tpr_diff + fpr_diff) / 2

def average_odds_error(y_true, y_pred, *, prot_attr=None, priv_group=None,
                       pos_label=1, sample_weight=None):
    r"""A relaxed version of equality of odds.

    Returns the average of the absolute difference in FPR and TPR for the
    unprivileged and privileged groups:

    .. math::

        \dfrac{|FPR_{D = \text{unprivileged}} - FPR_{D = \text{privileged}}|
        + |TPR_{D = \text{unprivileged}} - TPR_{D = \text{privileged}}|}{2}

    A value of 0 indicates equality of odds.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group. If
            prot_attr is binary, this may be ``None``.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Average odds error.
    """
    if priv_group is None:
        priv_group = check_groups(y_true, prot_attr=prot_attr,
                                  ensure_binary=True)[0][0]
    fpr_diff = -difference(specificity_score, y_true, y_pred,
                           prot_attr=prot_attr, priv_group=priv_group,
                           pos_label=pos_label, sample_weight=sample_weight)
    tpr_diff = difference(recall_score, y_true, y_pred, prot_attr=prot_attr,
                          priv_group=priv_group, pos_label=pos_label,
                          sample_weight=sample_weight)
    return (abs(tpr_diff) + abs(fpr_diff)) / 2

def average_predictive_value_difference(y_true, y_pred, *, prot_attr=None, priv_group=1,
                            pos_label=1, sample_weight=None):
    r"""Returns the average of the difference in positive predictive value and false omission rate for the unprivileged and privileged groups:

    .. math::

        \dfrac{(PPV_{D = \text{unprivileged}} - PPV_{D = \text{privileged}})
        + (FOR_{D = \text{unprivileged}} - FOR_{D = \text{privileged}})}{2}

    A value of 0 indicates equality of chance of success.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Average predictive value difference.
    """
    for_diff = difference(false_omission_rate_error, y_true, y_pred,
                           prot_attr=prot_attr, priv_group=priv_group,
                           pos_label=pos_label, sample_weight=sample_weight)
    ppv_diff = difference(precision_score, y_true, y_pred, prot_attr=prot_attr,
                          priv_group=priv_group, pos_label=pos_label,
                          sample_weight=sample_weight)
    return (ppv_diff + for_diff) / 2

def class_imbalance(y_true, y_pred=None, *, prot_attr=None, priv_group=1,
                    sample_weight=None):
    r"""Compute the class imbalance, :math:`\frac{N_u - N_p}{N_u + N_p}`.

    Where :math:`N_u` is the number of samples in the unprivileged group and
    :math:`N_p` is the number of samples in the privileged group.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like, optional): Estimated targets. Ignored.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar): The label of the privileged group.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Class imbalance.
    """
    diff = difference(num_samples, y_true, prot_attr=prot_attr,
                      priv_group=priv_group, sample_weight=sample_weight)
    return diff / num_samples(y_true, sample_weight=sample_weight)

def kl_divergence(y_true, y_pred=None, *, prot_attr=None, priv_group=1,
                  sample_weight=None):
    r"""Compute the Kullback-Leibler divergence, :math:`KL(P_p||P_u) = \sum_y
    P_p(y)\log\left(\frac{P_p(y)}{P_u(y)}\right)`

    where :math:`P_p` is the probability distribution over labels of the
    privileged group and, similiarly, :math:`P_u` is the distribution of the
    unprivileged group.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values. If y_pred
            is provided, this is ignored.
        y_pred (array-like, optional): Estimated targets as returned by a
            classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        priv_group (scalar): The label of the privileged group.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: KL divergence.
    """
    rate = base_rate if y_pred is None else selection_rate
    support = np.unique(y_true)  # TODO: is it correct to always use y_true?
    groups, _ = check_groups(y_true, prot_attr, ensure_binary=True)
    priv = np.unique(groups).tolist().index(priv_group)
    P1, P2 = zip(*[intersection(rate, y_true, y_pred, prot_attr=prot_attr,
                                pos_label=i, sample_weight=sample_weight)
                   for i in support])
    return sum(rel_entr(P1, P2)) if priv == 0 else sum(rel_entr(P2, P1))

def conditional_demographic_disparity(y_true, y_pred=None, *, prot_attr=None,
                                      pos_label=1, sample_weight=None):
    r"""Conditional demographic disparity, :math:`CDD = \frac{1}{\sum_i N_i}
    \sum_i N_i\cdot DD_i`

    where :math:`DD_i = \frac{N_{i, -}}{\sum_j N_{j, -}} - \frac{N_{i, +}}{
    \sum_j N_{j, +}}`.

    :math:`N_{i, +}` signifies the number of samples belonging to group
    :math:`i` that have favorable labels while :math:`N_{i, -}` signifies those
    that have negative labels [#watcher21]_.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values. If y_pred
            is provided, this is ignored.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        pos_label (scalar, optional): The label of the positive class.
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Conditional demographic disparity.

    References:
        .. [#watcher21] `S. Wachter, B. Mittelstadt, and C. Russell, "Why
           fairness cannot be automated: Bridging the gap between EU
           non-discrimination law and AI," Computer Law & Security Review,
           Volume 41, 2021. <https://doi.org/10.1016/j.clsr.2021.105567>`_
    """
    def dd(y_true, y_pred=None, pop_pos=1, pop_neg=1, pos_label=1,
           sample_weight=None):
        y_group = y_true if y_pred is None else y_pred
        y_pos, y_neg = num_pos_neg(y_group, pos_label=pos_label,
                                   sample_weight=sample_weight)
        return y_neg/pop_neg - y_pos/pop_pos

    pop_pos, pop_neg = num_pos_neg(y_true, y_pred, pos_label=pos_label,
                                   sample_weight=sample_weight)
    ddi = intersection(dd, y_true, y_pred, pop_pos=pop_pos, pop_neg=pop_neg,
                       prot_attr=prot_attr, pos_label=pos_label,
                       sample_weight=sample_weight)
    n = intersection(num_samples, y_true, prot_attr=prot_attr,
                     sample_weight=sample_weight)
    return np.dot(n, ddi) / sum(n)

# TODO: use soft scores if y is probas_pred
def smoothed_edf(y_true, y_pred=None, *, prot_attr=None, pos_label=1,
                 concentration=1.0, sample_weight=None):
    r"""Smoothed empirical differential fairness (EDF).

    .. math::
        e^{-\epsilon} \leq \frac{\sum_{A=s_i}{P(y|x)} + \alpha}{N_{s_i} + |R_Y|\alpha}
        \frac{N_{s_j} + |R_Y|\alpha}{\sum_{A=s_j}{P(y|x) + \alpha}} \leq e^\epsilon

    See [#foulds18]_ for more details.

    Note:
        If only y_true is provided, this will return the maximum epsilon for any
        two intersectional groups (smoothed EDF of the original dataset). If
        both y_true and y_pred are provided, only y_pred is used.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values. If y_pred
            is provided, this is ignored.
        y_pred (array-like, optional): Estimated targets as returned by a
            classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        pos_label (scalar, optional): The label of the positive class.
        concentration (scalar, optional): Dirichlet smoothing concentration
            parameter :math:`|R_Y|\alpha` (must be non-negative).
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Smoothed EDF, :math:`\epsilon`. Lower is better.

    See also:
        :func:`intersection`, :func:`smoothed_base_rate`

    References:
        .. [#foulds18] J. R. Foulds, R. Islam, K. N. Keya, and S. Pan,
           "An Intersectional Definition of Fairness," arXiv preprint
           arXiv:1807.08362, 2018.
    """
    rate = smoothed_base_rate if y_pred is None else smoothed_selection_rate
    sbr = intersection(rate, y_true, y_pred, prot_attr=prot_attr,
                       sample_weight=sample_weight, pos_label=pos_label,
                       concentration=concentration)

    logsbr = np.log(sbr)
    pos_ratio = max(abs(i - j) for i, j in permutations(logsbr, 2))
    lognegsbr = np.log(1 - np.array(sbr))
    neg_ratio = max(abs(i - j) for i, j in permutations(lognegsbr, 2))
    return max(pos_ratio, neg_ratio)

def df_bias_amplification(y_true, y_pred, *, prot_attr=None, pos_label=1,
                          concentration=1.0, sample_weight=None):
    r"""Differential fairness bias amplification.

    Measures the increase in unfairness attributable to a classifier compared to
    the original data. See [#foulds18]_ for more details.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, keyword-only): Protected attribute(s). If
            ``None``, all protected attributes in y_true are used.
        pos_label (scalar, optional): The label of the positive class.
        concentration (scalar, optional): Dirichlet smoothing concentration
            parameter :math:`|R_Y|\alpha` (must be non-negative).
        sample_weight (array-like, optional): Sample weights.

    Returns:
        float: Difference in smoothed EDF between the classifier and the
        original dataset, :math:`\epsilon_{\text{classifier}}
        - \epsilon_{\text{data}}`. Lower is better.

    References:
        .. [#foulds18] J. R. Foulds, R. Islam, K. N. Keya, and S. Pan,
           "An Intersectional Definition of Fairness," arXiv preprint
           arXiv:1807.08362, 2018.
    """
    eps_true = smoothed_edf(y_true, prot_attr=prot_attr, pos_label=pos_label,
                            concentration=concentration,
                            sample_weight=sample_weight)
    eps_pred = smoothed_edf(y_true, y_pred, prot_attr=prot_attr,
                            pos_label=pos_label, concentration=concentration,
                            sample_weight=sample_weight)
    return eps_pred - eps_true

def mdss_bias_score(y_true, probas_pred, X=None, subset=None, *, pos_label=1,
                    scoring='Bernoulli', overpredicted=True, penalty=1e-17,
                    **kwargs):
    """Compute the bias score for a prespecified group of records using a
    given scoring function.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        probas_pred (array-like): Probability estimates of the positive class.
        X (DataFrame, optional): The dataset (containing the features) that was
            used to predict `probas_pred`. If not specified, the subset is
            returned as indices.
        subset (dict, optional): Mapping of column names to list of values.
            Samples are included in the subset if they match any value in each
            of the columns provided. If `X` is not specified, `subset` may
            be of the form `{'index': [0, 1, ...]}` or `None`. If `None`, score
            over the full set (note: `penalty` is irrelevant in this case).
        pos_label (scalar, optional): Label of the positive class.
        scoring (str or class): One of 'Bernoulli' or 'BerkJones' or
            subclass of
            :class:`aif360.metrics.mdss.ScoringFunctions.ScoringFunction`.
        overpredicted (bool): Flag for which direction to scan: `True` means we
            scan for a group whose expectations/predictions are systematically
            higher than observed. In other words, we scan for a group whose
            observed is systematically lower than the expectations. `False`
            means we scan for a group whose expectations/predictions are
            systematically lower than observed (observed is systematically
            higher than the expectations).
        privileged (bool): Deprecated. Use overpredicted instead.
        penalty (scalar): Penalty coefficient. Should be positive. The higher
            the penalty, the less complex (number of features and feature
            values) the highest scoring subset that gets returned is.
        **kwargs: Additional kwargs to be passed to `scoring` (not including
            `direction`).

    Returns:
        float: Bias score for the given group.

    See also:
        :func:`mdss_bias_scan`
    """

    if X is None:
        X = pd.DataFrame({'index': range(len(y_true))})
    else:
        X = X.reset_index(drop=True)  # match all indices

    expected = pd.Series(probas_pred).reset_index(drop=True)
    outcomes = pd.Series(y_true == pos_label, dtype=int).reset_index(drop=True)

    # TODO: DEPRECATED. Remove in next version.
    if 'privileged' in kwargs:
        warnings.warn("privileged is deprecated. Use overpredicted instead.",
                      category=FutureWarning)
        overpredicted = kwargs['privileged']
    direction = 'negative' if overpredicted else 'positive'
    kwargs['direction'] = direction

    if scoring == 'Bernoulli':
        scoring_function = Bernoulli(**kwargs)
    elif scoring == 'BerkJones':
        scoring_function = BerkJones(**kwargs)
    else:
        scoring_function = scoring(**kwargs)
    scanner = MDSS(scoring_function)

    return scanner.score_current_subset(X, expected, outcomes, subset or {}, penalty)


# ========================== INDIVIDUAL FAIRNESS ===============================
def generalized_entropy_index(b, alpha=2):
    r"""Generalized entropy index measures inequality over a population.

    .. math::

        \mathcal{E}(\alpha) = \begin{cases}
            \frac{1}{n \alpha (\alpha-1)}\sum_{i=1}^n\left[\left(\frac{b_i}{\mu}\right)^\alpha - 1\right],& \alpha \ne 0, 1,\\
            \frac{1}{n}\sum_{i=1}^n\frac{b_{i}}{\mu}\ln\frac{b_{i}}{\mu},& \alpha=1,\\
            -\frac{1}{n}\sum_{i=1}^n\ln\frac{b_{i}}{\mu},& \alpha=0.
        \end{cases}

    Args:
        b (array-like): Parameter over which to calculate the entropy index.
        alpha (scalar): Parameter that regulates the weight given to distances
            between values at different parts of the distribution. A value of 0
            is equivalent to the mean log deviation, 1 is the Theil index, and 2
            is half the squared coefficient of variation.
    """
    if alpha == 0:
        return -(np.log(b / b.mean()) / b.mean()).mean()
    elif alpha == 1:
        # moving the b inside the log allows for 0 values
        return (np.log((b / b.mean())**b) / b.mean()).mean()
    else:
        return ((b / b.mean())**alpha - 1).mean() / (alpha * (alpha - 1))

def generalized_entropy_error(y_true, y_pred, alpha=2, pos_label=1):
    #                           sample_weight=None):
    r"""Compute the generalized entropy.

    Generalized entropy index is proposed as a unified individual and
    group fairness measure in [#speicher18]_.

    Uses :math:`b_i = \hat{y}_i - y_i + 1`. See
    :func:`generalized_entropy_index` for details.

    Args:
        y_true (array-like): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        alpha (scalar, optional): Parameter that regulates the weight given to
            distances between values at different parts of the distribution. A
            value of 0 is equivalent to the mean log deviation, 1 is the Theil
            index, and 2 is half the squared coefficient of variation.
        pos_label (scalar, optional): The label of the positive class.

    See also:
        :func:`generalized_entropy_index`

    References:
        .. [#speicher18] `T. Speicher, H. Heidari, N. Grgic-Hlaca,
           K. P. Gummadi, A. Singla, A. Weller, and M. B. Zafar, "A Unified
           Approach to Quantifying Algorithmic Unfairness: Measuring Individual
           and Group Unfairness via Inequality Indices," ACM SIGKDD
           International Conference on Knowledge Discovery and Data Mining,
           2018. <https://dl.acm.org/citation.cfm?id=3220046>`_
    """
    b = 1 + (y_pred == pos_label) - (y_true == pos_label)
    return generalized_entropy_index(b, alpha=alpha)

def between_group_generalized_entropy_error(y_true, y_pred, prot_attr=None,
        priv_group=None, alpha=2, pos_label=1):
    r"""Compute the between-group generalized entropy.

    Between-group generalized entropy index is proposed as a group
    fairness measure in [#speicher18]_ and is one of two terms that the
    generalized entropy index decomposes to.

    Args:
        y_true (pandas.Series): Ground truth (correct) target values.
        y_pred (array-like): Estimated targets as returned by a classifier.
        prot_attr (array-like, optional): Protected attribute(s). If ``None``,
            all protected attributes in y_true are used.
        priv_group (scalar, optional): The label of the privileged group. If
            provided, the index will be computed between only the privileged and
            unprivileged groups. Otherwise, the index will be computed between
            all groups defined by the prot_attr.
        alpha (scalar, optional): Parameter that regulates the weight given to
            distances between values at different parts of the distribution. A
            value of 0 is equivalent to the mean log deviation, 1 is the Theil
            index, and 2 is half the squared coefficient of variation.
        pos_label (scalar, optional): The label of the positive class.

    See also:
        :func:`generalized_entropy_index`

    References:
        .. [#speicher18] `T. Speicher, H. Heidari, N. Grgic-Hlaca,
           K. P. Gummadi, A. Singla, A. Weller, and M. B. Zafar, "A Unified
           Approach to Quantifying Algorithmic Unfairness: Measuring Individual
           and Group Unfairness via Inequality Indices," ACM SIGKDD
           International Conference on Knowledge Discovery and Data Mining,
           2018. <https://dl.acm.org/citation.cfm?id=3220046>`_
    """
    groups, _ = check_groups(y_true, prot_attr)
    b = np.empty_like(y_true, dtype='float')
    if priv_group is not None:
        groups = [1 if g == priv_group else 0 for g in groups]
    for g in np.unique(groups):
        b[groups == g] = (1 + (y_pred[groups == g] == pos_label)
                            - (y_true[groups == g] == pos_label)).mean()
    return generalized_entropy_index(b, alpha=alpha)

def theil_index(b):
    r"""The Theil index is the :func:`generalized_entropy_index` with
    :math:`\alpha = 1`.

    Args:
        b (array-like): Parameter over which to calculate the entropy index.

    See also:
        :func:`generalized_entropy_index`
    """
    return generalized_entropy_index(b, alpha=1)

def coefficient_of_variation(b):
    r"""The coefficient of variation is the square root of two times the
    :func:`generalized_entropy_index` with :math:`\alpha = 2`.

    Args:
        b (array-like): Parameter over which to calculate the entropy index.

    See also:
        :func:`generalized_entropy_index`
    """
    return np.sqrt(2 * generalized_entropy_index(b, alpha=2))


# TODO: use sample_weight?
def consistency_score(X, y, n_neighbors=5):
    r"""Compute the consistency score.

    Individual fairness metric from [#zemel13]_ that measures how similar the
    labels are for similar instances.

    .. math::
        1 - \frac{1}{n}\sum_{i=1}^n |\hat{y}_i -
        \frac{1}{\text{n_neighbors}} \sum_{j\in\mathcal{N}_{\text{n_neighbors}}(x_i)} \hat{y}_j|

    Args:
        X (array-like): Sample features.
        y (array-like): Sample targets.
        n_neighbors (int, optional): Number of neighbors for the knn
            computation.

    References:
        .. [#zemel13] `R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,
           "Learning Fair Representations," International Conference on Machine
           Learning, 2013. <http://proceedings.mlr.press/v28/zemel13.html>`_
    """
    # cast as ndarrays
    X, y = check_X_y(X, y)
    # learn a KNN on the features
    nbrs = NearestNeighbors(n_neighbors=n_neighbors, algorithm='ball_tree')
    nbrs.fit(X)
    indices = nbrs.kneighbors(X, return_distance=False)

    # compute consistency score
    return 1 - abs(y - y[indices].mean(axis=1)).mean()


# ================================ ALIASES =====================================
def sensitivity_score(y_true, y_pred, pos_label=1, sample_weight=None):
    """Alias of :func:`sklearn.metrics.recall_score` for binary classes only."""
    return recall_score(y_true, y_pred, pos_label=pos_label,
                        sample_weight=sample_weight)

def false_negative_rate_error(y_true, y_pred, pos_label=1, sample_weight=None):
    return 1 - recall_score(y_true, y_pred, pos_label=pos_label,
                            sample_weight=sample_weight)

def false_positive_rate_error(y_true, y_pred, pos_label=1, sample_weight=None):
    return 1 - specificity_score(y_true, y_pred, pos_label=pos_label,
                                 sample_weight=sample_weight)

def mean_difference(y_true, y_pred=None, *, prot_attr=None, priv_group=1,
                    pos_label=1, sample_weight=None):
    """Alias of :func:`statistical_parity_difference`."""
    return statistical_parity_difference(y_true, y_pred, prot_attr=prot_attr,
            priv_group=priv_group, pos_label=pos_label,
            sample_weight=sample_weight)



# Adapted from:
# http://www.christianmoscardi.com/blog/2016/01/20/jupyter-testing.html

from .notebook_runner import notebook_run
import os


def test_reject_option_classification():
    nb, errors = notebook_run(os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            '..', 'examples', 'demo_reject_option_classification.ipynb'))

    if len(errors) > 0:
        for err in errors:
            for tbi in err['traceback']:
                print(tbi)
        raise AssertionError("errors in notebook testcases")



import numpy as np

from aif360.datasets import StructuredDataset
'''
Multiclass supports the multiple values in the favorable and unfavorable label's
'''
class MulticlassLabelDataset(StructuredDataset):
    """Base class for all structured datasets with multiclass labels."""

    def __init__(self, favorable_label = [1.], unfavorable_label = [0.], **kwargs):
        """
        Args:
            favorable_label (list): Label value which is considered favorable
                (i.e. "positive").
            unfavorable_label (list): Label value which is considered
                unfavorable (i.e. "negative").
            **kwargs: StructuredDataset arguments.
        """
        self.favorable_label = favorable_label
        self.unfavorable_label = unfavorable_label

        super(MulticlassLabelDataset, self).__init__(**kwargs)
        

    def validate_dataset(self):
        """Error checking and type validation.
​
        Raises:
            ValueError: `labels` must be shape [n, 1].
            ValueError: `favorable_label` and `unfavorable_label` must be the
                only values present in `labels`.
        """
        # fix scores before validating
        if np.all(self.scores == self.labels):
            for i in range(0,len(self.scores)):
                if self.scores[i] in self.favorable_label:
                    self.scores[i] = float(1)
                else:
                    self.scores[i] = float(0)

        super(MulticlassLabelDataset, self).validate_dataset()

        # =========================== SHAPE CHECKING ===========================
        # Verify if the labels are only 1 column
        if self.labels.shape[1] != 1:
            raise ValueError("MulticlassLabelDataset only supports single-column "
                "labels:\n\tlabels.shape = {}".format(self.labels.shape))

        # =========================== VALUE CHECKING ===========================
        # Check if the favorable and unfavorable labels match those in the dataset
        if (not set(self.labels.ravel()) <=
                set(self.favorable_label + (self.unfavorable_label))):
            raise ValueError("The favorable and unfavorable labels provided do "
                             "not match the labels in the dataset.")



from abc import ABC, abstractmethod
import copy


class Dataset(ABC):
    """Abstract base class for datasets."""

    @abstractmethod
    def __init__(self, **kwargs):
        self.metadata = kwargs.pop('metadata', dict()) or dict()
        self.metadata.update({
            'transformer': '{}.__init__'.format(type(self).__name__),
            'params': kwargs,
            'previous': []
        })
        self.validate_dataset()

    def validate_dataset(self):
        """Error checking and type validation."""
        pass

    def copy(self, deepcopy=False):
        """Convenience method to return a copy of this dataset.

        Args:
            deepcopy (bool, optional): :func:`~copy.deepcopy` this dataset if
                `True`, shallow copy otherwise.

        Returns:
            Dataset: A new dataset with fields copied from this object and
            metadata set accordingly.
        """
        cpy = copy.deepcopy(self) if deepcopy else copy.copy(self)
        # preserve any user-created fields
        cpy.metadata = cpy.metadata.copy()
        cpy.metadata.update({
            'transformer': '{}.copy'.format(type(self).__name__),
            'params': {'deepcopy': deepcopy},
            'previous': [self]
        })
        return cpy

    @abstractmethod
    def export_dataset(self):
        """Save this Dataset to disk."""
        raise NotImplementedError

    @abstractmethod
    def split(self, num_or_size_splits, shuffle=False):
        """Split this dataset into multiple partitions.

        Args:
            num_or_size_splits (array or int): If `num_or_size_splits` is an
                int, *k*, the value is the number of equal-sized folds to make
                (if *k* does not evenly divide the dataset these folds are
                approximately equal-sized). If `num_or_size_splits` is an array
                of type int, the values are taken as the indices at which to
                split the dataset. If the values are floats (< 1.), they are
                considered to be fractional proportions of the dataset at which
                to split.
            shuffle (bool, optional): Randomly shuffle the dataset before
                splitting.

        Returns:
            list(Dataset): Splits. Contains *k* or `len(num_or_size_splits) + 1`
            datasets depending on `num_or_size_splits`.
        """
        raise NotImplementedError



from aif360.datasets import AdultDataset, GermanDataset, CompasDataset
import pandas as pd
import numpy as np

def load_preproc_data_adult(protected_attributes=None, sub_samp=False, balance=False):
    def custom_preprocessing(df):
        """The custom pre-processing function is adapted from
            https://github.com/fair-preprocessing/nips2017/blob/master/Adult/code/Generate_Adult_Data.ipynb
            If sub_samp != False, then return smaller version of dataset truncated to tiny_test data points.
        """

        # Group age by decade
        df['Age (decade)'] = df['age'].apply(lambda x: x//10*10)
        # df['Age (decade)'] = df['age'].apply(lambda x: np.floor(x/10.0)*10.0)

        def group_edu(x):
            if x <= 5:
                return '<6'
            elif x >= 13:
                return '>12'
            else:
                return x

        def age_cut(x):
            if x >= 70:
                return '>=70'
            else:
                return x

        def group_race(x):
            if x == "White":
                return 1.0
            else:
                return 0.0

        # Cluster education and age attributes.
        # Limit education range
        df['Education Years'] = df['education-num'].apply(lambda x: group_edu(x))
        df['Education Years'] = df['Education Years'].astype('category')

        # Limit age range
        df['Age (decade)'] = df['Age (decade)'].apply(lambda x: age_cut(x))

        # Rename income variable
        df['Income Binary'] = df['income-per-year']
        df['Income Binary'] = df['Income Binary'].replace(to_replace='>50K.', value='>50K', regex=True)
        df['Income Binary'] = df['Income Binary'].replace(to_replace='<=50K.', value='<=50K', regex=True)

        # Recode sex and race
        df['sex'] = df['sex'].replace({'Female': 0.0, 'Male': 1.0})
        df['race'] = df['race'].apply(lambda x: group_race(x))

        if sub_samp and not balance:
            df = df.sample(sub_samp)
        if sub_samp and balance:
            df_0 = df[df['Income Binary'] == '<=50K']
            df_1 = df[df['Income Binary'] == '>50K']
            df_0 = df_0.sample(int(sub_samp/2))
            df_1 = df_1.sample(int(sub_samp/2))
            df = pd.concat([df_0, df_1])
        return df

    XD_features = ['Age (decade)', 'Education Years', 'sex', 'race']
    D_features = ['sex', 'race'] if protected_attributes is None else protected_attributes
    Y_features = ['Income Binary']
    X_features = list(set(XD_features)-set(D_features))
    categorical_features = ['Age (decade)', 'Education Years']

    # privileged classes
    all_privileged_classes = {"sex": [1.0],
                              "race": [1.0]}

    # protected attribute maps
    all_protected_attribute_maps = {"sex": {1.0: 'Male', 0.0: 'Female'},
                                    "race": {1.0: 'White', 0.0: 'Non-white'}}

    return AdultDataset(
        label_name=Y_features[0],
        favorable_classes=['>50K', '>50K.'],
        protected_attribute_names=D_features,
        privileged_classes=[all_privileged_classes[x] for x in D_features],
        instance_weights_name=None,
        categorical_features=categorical_features,
        features_to_keep=X_features+Y_features+D_features,
        na_values=['?'],
        metadata={'label_maps': [{1.0: '>50K', 0.0: '<=50K'}],
                  'protected_attribute_maps': [all_protected_attribute_maps[x]
                                for x in D_features]},
        custom_preprocessing=custom_preprocessing)

def load_preproc_data_compas(protected_attributes=None):
    def custom_preprocessing(df):
        """The custom pre-processing function is adapted from
            https://github.com/fair-preprocessing/nips2017/blob/master/compas/code/Generate_Compas_Data.ipynb
        """

        df = df[['age', 'c_charge_degree', 'race', 'age_cat', 'score_text',
                 'sex', 'priors_count', 'days_b_screening_arrest', 'decile_score',
                 'is_recid', 'two_year_recid', 'c_jail_in', 'c_jail_out']]

        # Indices of data samples to keep
        ix = df['days_b_screening_arrest'] <= 30
        ix = (df['days_b_screening_arrest'] >= -30) & ix
        ix = (df['is_recid'] != -1) & ix
        ix = (df['c_charge_degree'] != "O") & ix
        ix = (df['score_text'] != 'N/A') & ix
        df = df.loc[ix,:]
        df['length_of_stay'] = (pd.to_datetime(df['c_jail_out'])-
                                pd.to_datetime(df['c_jail_in'])).apply(
                                                        lambda x: x.days)

        # Restrict races to African-American and Caucasian
        dfcut = df.loc[~df['race'].isin(['Native American','Hispanic','Asian','Other']),:]

        # Restrict the features to use
        dfcutQ = dfcut[['sex','race','age_cat','c_charge_degree','score_text','priors_count','is_recid',
                'two_year_recid','length_of_stay']].copy()

        # Quantize priors count between 0, 1-3, and >3
        def quantizePrior(x):
            if x <=0:
                return '0'
            elif 1<=x<=3:
                return '1 to 3'
            else:
                return 'More than 3'

        # Quantize length of stay
        def quantizeLOS(x):
            if x<= 7:
                return '<week'
            if 8<x<=93:
                return '<3months'
            else:
                return '>3 months'

        # Quantize length of stay
        def adjustAge(x):
            if x == '25 - 45':
                return '25 to 45'
            else:
                return x

        # Quantize score_text to MediumHigh
        def quantizeScore(x):
            if (x == 'High')| (x == 'Medium'):
                return 'MediumHigh'
            else:
                return x

        def group_race(x):
            if x == "Caucasian":
                return 1.0
            else:
                return 0.0

        dfcutQ['priors_count'] = dfcutQ['priors_count'].apply(lambda x: quantizePrior(x))
        dfcutQ['length_of_stay'] = dfcutQ['length_of_stay'].apply(lambda x: quantizeLOS(x))
        dfcutQ['score_text'] = dfcutQ['score_text'].apply(lambda x: quantizeScore(x))
        dfcutQ['age_cat'] = dfcutQ['age_cat'].apply(lambda x: adjustAge(x))

        # Recode sex and race
        dfcutQ['sex'] = dfcutQ['sex'].replace({'Female': 1.0, 'Male': 0.0})
        dfcutQ['race'] = dfcutQ['race'].apply(lambda x: group_race(x))

        features = ['two_year_recid',
                    'sex', 'race',
                    'age_cat', 'priors_count', 'c_charge_degree']

        # Pass vallue to df
        df = dfcutQ[features]

        return df

    XD_features = ['age_cat', 'c_charge_degree', 'priors_count', 'sex', 'race']
    D_features = ['sex', 'race']  if protected_attributes is None else protected_attributes
    Y_features = ['two_year_recid']
    X_features = list(set(XD_features)-set(D_features))
    categorical_features = ['age_cat', 'priors_count', 'c_charge_degree']

    # privileged classes
    all_privileged_classes = {"sex": [1.0],
                              "race": [1.0]}

    # protected attribute maps
    all_protected_attribute_maps = {"sex": {0.0: 'Male', 1.0: 'Female'},
                                    "race": {1.0: 'Caucasian', 0.0: 'Not Caucasian'}}


    return CompasDataset(
        label_name=Y_features[0],
        favorable_classes=[0],
        protected_attribute_names=D_features,
        privileged_classes=[all_privileged_classes[x] for x in D_features],
        instance_weights_name=None,
        categorical_features=categorical_features,
        features_to_keep=X_features+Y_features+D_features,
        na_values=[],
        metadata={'label_maps': [{1.0: 'Did recid.', 0.0: 'No recid.'}],
                  'protected_attribute_maps': [all_protected_attribute_maps[x]
                                for x in D_features]},
        custom_preprocessing=custom_preprocessing)

def load_preproc_data_german(protected_attributes=None):
    """
    Load and pre-process german credit dataset.
    Args:
        protected_attributes(list or None): If None use all possible protected
            attributes, else subset the protected attributes to the list.

    Returns:
        GermanDataset: An instance of GermanDataset with required pre-processing.

    """
    def custom_preprocessing(df):
        """ Custom pre-processing for German Credit Data
        """

        def group_credit_hist(x):
            if x in ['A30', 'A31', 'A32']:
                return 'None/Paid'
            elif x == 'A33':
                return 'Delay'
            elif x == 'A34':
                return 'Other'
            else:
                return 'NA'

        def group_employ(x):
            if x == 'A71':
                return 'Unemployed'
            elif x in ['A72', 'A73']:
                return '1-4 years'
            elif x in ['A74', 'A75']:
                return '4+ years'
            else:
                return 'NA'

        def group_savings(x):
            if x in ['A61', 'A62']:
                return '<500'
            elif x in ['A63', 'A64']:
                return '500+'
            elif x == 'A65':
                return 'Unknown/None'
            else:
                return 'NA'

        def group_status(x):
            if x in ['A11', 'A12']:
                return '<200'
            elif x in ['A13']:
                return '200+'
            elif x == 'A14':
                return 'None'
            else:
                return 'NA'

        status_map = {'A91': 1.0, 'A93': 1.0, 'A94': 1.0,
                    'A92': 0.0, 'A95': 0.0}
        df['sex'] = df['personal_status'].replace(status_map)


        # group credit history, savings, and employment
        df['credit_history'] = df['credit_history'].apply(lambda x: group_credit_hist(x))
        df['savings'] = df['savings'].apply(lambda x: group_savings(x))
        df['employment'] = df['employment'].apply(lambda x: group_employ(x))
        df['age'] = (df['age'] >= 26).astype(float)
        df['status'] = df['status'].apply(lambda x: group_status(x))

        return df

    # Feature partitions
    XD_features = ['credit_history', 'savings', 'employment', 'sex', 'age']
    D_features = ['sex', 'age'] if protected_attributes is None else protected_attributes
    Y_features = ['credit']
    X_features = list(set(XD_features)-set(D_features))
    categorical_features = ['credit_history', 'savings', 'employment']

    # privileged classes
    all_privileged_classes = {"sex": [1.0],
                              "age": [1.0]}

    # protected attribute maps
    all_protected_attribute_maps = {"sex": {1.0: 'Male', 0.0: 'Female'},
                                    "age": {1.0: 'Old', 0.0: 'Young'}}

    return GermanDataset(
        label_name=Y_features[0],
        favorable_classes=[1],
        protected_attribute_names=D_features,
        privileged_classes=[all_privileged_classes[x] for x in D_features],
        instance_weights_name=None,
        categorical_features=categorical_features,
        features_to_keep=X_features+Y_features+D_features,
        metadata={ 'label_maps': [{1.0: 'Good Credit', 2.0: 'Bad Credit'}],
                   'protected_attribute_maps': [all_protected_attribute_maps[x]
                                for x in D_features]},
        custom_preprocessing=custom_preprocessing)



"""File containing common pytest fixtures used by multiple test files."""
import numpy as np
import pytest

from aif360.datasets import GermanDataset, AdultDataset
from aif360.sklearn.datasets import fetch_german, fetch_adult

@pytest.fixture(scope='module')
def old_german():
    german = GermanDataset(categorical_features=['foreign_worker'],
            features_to_keep=['month', 'credit_amount',
            'investment_as_income_percentage', 'residence_since', 'age',
            'number_of_credits', 'people_liable_for', 'sex'])
    german.features = np.concatenate(
            (german.features[:, :-3], german.features[:, -2:-4:-1]), axis=1)
    german.feature_names = german.feature_names[:-3] + ['foreign_worker', 'sex']
    german.labels[german.labels == 2] = 0
    return german

@pytest.fixture(scope='module')
def old_adult():
    return AdultDataset(instance_weights_name='fnlwgt', categorical_features=[],
            features_to_keep=['age', 'education-num', 'capital-gain',
                              'capital-loss', 'hours-per-week'],
            features_to_drop=[])


@pytest.fixture(scope='module')
def new_german():
    german = fetch_german(numeric_only=True)
    german.X.age = german.X.age.apply(lambda a: 1 if a > 25 else 0)
    return german

@pytest.fixture(scope='module')
def new_adult():
    return fetch_adult(numeric_only=True)



# Original work Copyright 2017 Flavio Calmon
# Modified work Copyright 2018 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
from warnings import warn

import numpy as np
import pandas as pd

from aif360.algorithms import Transformer
from aif360.datasets import BinaryLabelDataset


class OptimPreproc(Transformer):
    """Optimized preprocessing is a preprocessing technique that learns a
    probabilistic transformation that edits the features and labels in the data
    with group fairness, individual distortion, and data fidelity constraints
    and objectives [3]_.

    References:
        .. [3] F. P. Calmon, D. Wei, B. Vinzamuri, K. Natesan Ramamurthy, and
           K. R. Varshney. "Optimized Pre-Processing for Discrimination
           Prevention." Conference on Neural Information Processing Systems,
           2017.

    Based on code available at: https://github.com/fair-preprocessing/nips2017
    """

    def __init__(self, optimizer, optim_options, unprivileged_groups=None,
                 privileged_groups=None, verbose=False, seed=None):
        """
        Args:
            optimizer (class): Optimizer class.
            optim_options (dict): Options for optimization to estimate the
                transformation.
            unprivileged_groups (dict): Representation for unprivileged group.
            privileged_groups (dict): Representation for privileged group.
            verbose (bool, optional): Verbosity flag for optimization.
            seed (int, optional): Seed to make `fit` and `predict` repeatable.

        Note:
            This algorithm does not use the privileged and unprivileged groups
            that are specified during initialization yet. Instead, it
            automatically attempts to reduce statistical parity difference
            between all possible combinations of groups in the dataset.
        """

        super(OptimPreproc, self).__init__(optimizer=optimizer,
            optim_options=optim_options,
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups, verbose=verbose, seed=seed)

        self.seed = seed
        self.optimizer = optimizer
        self.optim_options = optim_options
        self.verbose = verbose

        self.unprivileged_groups = unprivileged_groups
        self.privileged_groups = privileged_groups
        if unprivileged_groups or privileged_groups:
            warn("Privileged and unprivileged groups specified will not be "
                 "used. The protected attributes are directly specified in the "
                 "data preprocessing function. The current implementation "
                 "automatically adjusts for discrimination across all groups. "
                 "This can be changed by changing the optimization code.")

    def fit(self, dataset, sep='='):
        """Compute optimal pre-processing transformation based on distortion
        constraint.

        Args:
            dataset (BinaryLabelDataset): Dataset containing true labels.
            sep (str, optional): Separator for converting one-hot labels to
                categorical.
        Returns:
            OptimPreproc: Returns self.
        """
        if len(np.unique(dataset.instance_weights)) > 1:
            warn("Optimized pre-processing will ignore instance_weights in "
                 "the dataset during fit.")
        # Convert the dataset to a dataframe and preprocess
        df, _ = dataset.convert_to_dataframe(de_dummy_code=True, sep=sep,
                                             set_category=True)

        # Subset the protected attribute names and attribute values from
        # input parameters
        self.protected_attribute_names = dataset.protected_attribute_names
        self.privileged_protected_attributes = dataset.privileged_protected_attributes
        self.unprivileged_protected_attributes = dataset.unprivileged_protected_attributes

        # Feature names
        self.Y_feature_names = dataset.label_names
        self.X_feature_names = [n for n in df.columns.tolist()
                                if n not in self.Y_feature_names
                                and n not in self.protected_attribute_names]
        self.feature_names = (self.X_feature_names + self.Y_feature_names
                            + self.protected_attribute_names)

        # initialize a new OptTools object
        self.OpT = self.optimizer(df=df, features=self.feature_names)

        # Set features
        self.OpT.set_features(D=self.protected_attribute_names,
                              X=self.X_feature_names,
                              Y=self.Y_feature_names)

        # Set Distortion
        self.OpT.set_distortion(self.optim_options['distortion_fun'],
                                clist=self.optim_options['clist'])

        # solve optimization for previous parameters
        self.OpT.optimize(epsilon=self.optim_options['epsilon'],
                          dlist=self.optim_options['dlist'],
                          verbose=self.verbose)

        # Compute marginals
        self.OpT.compute_marginals()

        return self

    def transform(self, dataset, sep='=', transform_Y=True):
        """Transform the dataset to a new dataset based on the estimated
        transformation.

        Args:
            dataset (BinaryLabelDataset): Dataset containing labels that needs
                to be transformed.
            transform_Y (bool): Flag that mandates transformation of Y (labels).
        """

        if len(np.unique(dataset.instance_weights)) > 1:
            warn("Optimized pre-processing will ignore instance_weights in "
                 "the dataset during predict. The transformed dataset will "
                 "have all instance weights set to 1.")

        # Convert the dataset to a dataframe and preprocess
        df, _ = dataset.convert_to_dataframe(de_dummy_code=True, sep=sep,
                                             set_category=True)

        # Feature names
        Y_feature_names = dataset.label_names
        D_feature_names = self.protected_attribute_names
        X_feature_names = [n for n in df.columns.tolist()
                           if n not in self.Y_feature_names
                           and n not in self.protected_attribute_names]

        if (X_feature_names != self.X_feature_names or
            D_feature_names != self.protected_attribute_names):

           raise ValueError("The feature names of inputs and protected "
                            "attributes must match with the training dataset.")

        if transform_Y and (Y_feature_names != self.Y_feature_names):
            raise ValueError("The label names must match with that in the training dataset")

        if transform_Y:
            # randomized mapping when Y is requested to be transformed
            dfP_withY = self.OpT.dfP.applymap(lambda x: 0 if x < 1e-8 else x)
            dfP_withY = dfP_withY.divide(dfP_withY.sum(axis=1), axis=0)

            df_transformed = _apply_randomized_mapping(df, dfP_withY,
                features=D_feature_names+X_feature_names+Y_feature_names,
                random_seed=self.seed)
        else:
            # randomized mapping when Y is not requested to be transformed
            d1 = self.OpT.dfFull.reset_index().groupby(
                D_feature_names+X_feature_names).sum()
            d2 = d1.transpose().reset_index().groupby(X_feature_names).sum()
            dfP_noY = d2.transpose()
            dfP_noY = dfP_noY.drop(Y_feature_names, 1)
            dfP_noY = dfP_noY.applymap(lambda x: x if x > 1e-8 else 0)
            dfP_noY = dfP_noY/dfP_noY.sum()

            dfP_noY = dfP_noY.divide(dfP_noY.sum(axis=1), axis=0)

            df_transformed = _apply_randomized_mapping(
                                df, dfP_noY,
                                features=D_feature_names+X_feature_names,
                                random_seed=self.seed)

        # Map the protected attributes to numeric values
        for idx, p in enumerate(self.protected_attribute_names):
            pmap = dataset.metadata["protected_attribute_maps"][idx]
            pmap_rev = dict(zip(pmap.values(), pmap.keys()))
            df_transformed[p] = df_transformed[p].replace(pmap_rev)

        # Map the labels to numeric values
        for idx, p in enumerate(Y_feature_names):
            pmap = dataset.metadata["label_maps"][idx]
            pmap_rev = dict(zip(pmap.values(), pmap.keys()))
            df_transformed[p] = df_transformed[p].replace(pmap_rev)

        # Dummy code and convert to a dataset
        df_dum = pd.concat([pd.get_dummies(df_transformed.loc[:, X_feature_names],
                            prefix_sep="="),
                            df_transformed.loc[:, Y_feature_names+D_feature_names]],
                            axis=1)

        # Create a dataset out of df_dum
        dataset_transformed = BinaryLabelDataset(
            df=df_dum,
            label_names=Y_feature_names,
            protected_attribute_names=self.protected_attribute_names,
            privileged_protected_attributes=self.privileged_protected_attributes,
            unprivileged_protected_attributes=self.unprivileged_protected_attributes,
            favorable_label=dataset.favorable_label,
            unfavorable_label=dataset.unfavorable_label,
            metadata=dataset.metadata)

        return dataset_transformed

    def fit_transform(self, dataset, sep='=', transform_Y=True):
        """Perfom :meth:`fit` and :meth:`transform` sequentially."""

        return self.fit(dataset, sep=sep).transform(dataset, sep=sep,
                                                    transform_Y=transform_Y)

##############################
#### Supporting functions ####
##############################
def _apply_randomized_mapping(df, dfMap,
                              features=[], random_seed=None):
    """Apply Randomized mapping to create a new dataframe

    Args:
        df (DataFrame): Input dataframe
        dfMap (DataFrame): Mapping parameters
        features (list): Feature names for which the mapping needs to be applied
        random_seed (int): Random seed

    Returns:
        Perturbed version of df according to the randomizedmapping
    """

    if random_seed is not None:
        np.random.seed(seed=random_seed)

    df2 = df[features].copy()
    rem_cols = [l for l in df.columns
                if l not in features]
    if rem_cols != []:
        df3 = df[rem_cols].copy()

    idx_list = [tuple(i) for i in df2.itertuples(index=False)]

    draw_probs = dfMap.loc[idx_list]
    draws_possible = draw_probs.columns.tolist()

    # Make random draws - as part of randomizing transformation
    def draw_ind(x): return np.random.choice(range(len(draws_possible)), p=x)

    draw_inds = [draw_ind(x) for x in draw_probs.values]

    df2.loc[:, dfMap.columns.names] = [draws_possible[x] for x in draw_inds]

    if rem_cols != []:
        return pd.concat([df2, df3], axis=1)
    else:
        return df2



import unittest
import pandas as pd
from aif360.datasets import RegressionDataset
from aif360.algorithms.postprocessing.deterministic_reranking import DeterministicReranking

dataset = RegressionDataset(pd.DataFrame([
    ['r', 100],
    ['r', 90],
    ['r', 80],
    ['r', 70],
    ['b', 60],
    ['b', 50],
    ['b', 40],
    ['b', 30],
    ['b', 20],
    ['r', 10],
], columns=['s', 'score']), dep_var_name='score', protected_attribute_names=['s'], privileged_classes=[['r']])

class TestInputValidation(unittest.TestCase):
    def test_one_group(self):
        with self.assertRaises(Exception):
            d = DeterministicReranking([{'a': 0}, {'b': 0}], [{'a': 0}, {'b': 0}])
            d.fit(dataset)
    def test_diff_attr_names(self):
        with self.assertRaises(Exception):
            d = DeterministicReranking([{'a': 0}], [{'b': 0}])
            d.fit(dataset)
    def test_rec_size(self):
        with self.assertRaises(Exception):
            d = DeterministicReranking([{'a': 0}], [{'b': 0}])
            d.fit(dataset)
            d.predict(dataset, rec_size=-1, target_prop=[0.5, 0.5])
    def test_prop_len(self):
        with self.assertRaises(Exception):
            d = DeterministicReranking([{'a': 0}], [{'b': 0}])
            d.fit(dataset)
            d.predict(dataset, rec_size=1, target_prop=[0.5])

class TestValues(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        
        self.d = DeterministicReranking(privileged_groups=[{'s': 1}], unprivileged_groups=[{'s': 0}])
        self.d.fit(dataset)
        super().__init__(methodName)

    def test_wrong_type(self):
        with self.assertRaises(ValueError):
            self.d.predict(
                dataset, rec_size=6, target_prop=[0.5, 0.5], rerank_type='WRONG').convert_to_dataframe()[0]

    def test_greedy(self):
        ds = self.d.predict(
            dataset, rec_size=6, target_prop=[0.5, 0.5], rerank_type='Greedy').convert_to_dataframe()[0]
        actual = len(ds[ds['s'] == 1])/len(ds)
        expected = 0.5
        assert actual == expected

    def test_conserv(self):
        ds = self.d.predict(
            dataset, rec_size=6, target_prop=[0.5, 0.5], rerank_type='Conservative').convert_to_dataframe()[0]
        actual = len(ds[ds['s'] == 1])/len(ds)
        expected = 0.5
        assert actual == expected
    
    def test_relaxed(self):
        ds = self.d.predict(
            dataset, rec_size=6, target_prop=[0.5, 0.5], rerank_type='Relaxed').convert_to_dataframe()[0]
        actual = len(ds[ds['s'] == 1])/len(ds)
        expected = 0.5
        assert actual == expected

    def test_constrained(self):
        ds = self.d.predict(
            dataset, rec_size=6, target_prop=[0.5, 0.5], rerank_type='Constrained').convert_to_dataframe()[0]
        actual = len(ds[ds['s'] == 1])/len(ds)
        expected = 0.5
        assert actual == expected

if __name__ == '__main__':
    unittest.main()


import numpy as np
import pandas as pd

pd.set_option('display.max_rows', 10)
pd.set_option('display.max_columns', 10)
pd.set_option('display.width', 200)

from aif360.datasets import StructuredDataset

data = np.arange(12).reshape((4, 3))
cols = ['one', 'two', 'three', 'label']
labs = np.ones((4, 1))

df = pd.DataFrame(data=np.concatenate((data, labs), axis=1), columns=cols)

def test_temporarily_ignore():
    sd = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['one', 'three'])
    modified = sd.copy()
    modified.labels = sd.labels + 1
    assert sd != modified
    with sd.temporarily_ignore('labels'):
        assert sd == modified
    assert 'labels' not in sd.ignore_fields

def test_split():
    sd = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['two'])
    train, test = sd.split([0.5])
    train2, test2 = sd.split(2)

    assert train == train2
    assert test == test2
    assert np.all(np.concatenate((train.features, test.features)) == sd.features)

def test_k_folds():
    sd = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['two'])
    folds = sd.split(4)

    assert len(folds) == 4
    assert all(f.features.shape[0] == f.labels.shape[0]
            == f.protected_attributes.shape[0] == len(f.instance_names)
            == f.instance_weights.shape[0] == 1 for f in folds)

    folds = sd.split(3)
    assert folds[0].features.shape[0] == 2

def test_copy():
    sd = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['two'])
    sd2 = sd.copy()
    sd3 = sd.copy(True)

    sd.features[0] = 999
    assert np.all(sd2.features[0] == 999)
    assert not np.any(sd3.features[0] == 999)

def test_eq():
    sd = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['two'])
    sd2 = sd.copy()
    sd3 = sd.copy(True)
    sd4 = StructuredDataset(df=df, label_names=['label'], protected_attribute_names=['one', 'three'])

    assert sd == sd2
    assert sd == sd3
    assert sd2 == sd3
    assert sd != sd4



# Copyright 2019 Seth V. Neel, Michael J. Kearns, Aaron L. Roth, Zhiwei Steven Wu
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.
"""Class Auditor and Class Group implementing auditing for rich subgroup fairness in [KRNW18].

This module contains functionality to Audit an arbitrary classifier with respect to rich subgroup fairness,
where rich subgroup fairness is defined by hyperplanes over the sensitive attributes.

Basic Usage:
    auditor = Auditor(data_set, 'FP')
    # returns mean(predictions | y = 0) if 'FP' 1-mean(predictions | y = 1) if FN
    metric_baseline = auditor.get_baseline(y, predictions)
    group = auditor.get_group(dataset_yhat.labels, metric_baseline)
"""

import numpy as np
import pandas as pd
from sklearn import linear_model
from aif360.algorithms.inprocessing.gerryfair.reg_oracle_class import RegOracle
from aif360.algorithms.inprocessing.gerryfair import clean


class Group(object):
    """Group class: created by Auditor when identifying violation."""
    def __init__(self, func, group_size, weighted_disparity, disparity,
                 disparity_direction, group_rate):
        """Constructor for Group Class.

        :param func: the linear function that defines the group
        :param group_size: the proportion of the dataset in the group
        :param weighted_disparity: group_size*FP or FN disparity
        :param disparity: FN or FP disparity (absolute value)
        :param disparity_direction: indicator whether fp in group > fp_baseline, returns {1, -1}
        :param group_rate: FN or FN rate in the group
        """
        super(Group, self).__init__()
        self.func = func
        self.group_size = group_size
        self.weighted_disparity = weighted_disparity
        self.disparity = disparity
        self.disparity_direction = disparity_direction
        self.group_rate = group_rate

    def return_f(self):
        return [
            self.func, self.group_size, self.weighted_disparity,
            self.disparity, self.disparity_direction, self.group_rate
        ]


class Auditor:
    """This is the Auditor class. It is used in the training algorithm to repeatedly find subgroups that break the
    fairness disparity constraint. You can also use it independently as a stand alone auditor."""
    def __init__(self, dataset, fairness_def):
        """Auditor constructor.

        Args:
            :param dataset: dataset object subclassing StandardDataset.
            :param fairness_def: 'FP' or 'FN'
        """
        X, X_prime, y = clean.extract_df_from_ds(dataset)
        self.X_prime = X_prime
        self.y_input = y
        self.y_inverse = np.array(
            [abs(1 - y_value) for y_value in self.y_input])
        self.fairness_def = fairness_def
        if self.fairness_def not in ['FP', 'FN']:
            raise Exception(
                'Invalid fairness metric specified: {}. Please choose \'FP\' or \'FN\'.'
                .format(self.fairness_def))
        self.y = self.y_input
        # flip the labels for FN rate auditing
        if self.fairness_def == 'FN':
            self.y = self.y_inverse
        self.X_prime_0 = pd.DataFrame(
            [self.X_prime.iloc[u, :] for u, s in enumerate(self.y) if s == 0])

    def initialize_costs(self, n):
        """Initialize the costs for CSC problem that corresponds to auditing. See paper for details.

        Args:
            :param self: object of class Auditor
            :param n: size of the dataset

        Return:
            :return The costs for labeling a point 0, for labeling a point 1, as tuples.
        """
        costs_0 = None
        costs_1 = None
        if self.fairness_def == 'FP':
            costs_0 = [0.0] * n
            costs_1 = [-1.0 / n * (2 * i - 1) for i in self.y_input]

        elif self.fairness_def == 'FN':
            costs_1 = [0.0] * n
            costs_0 = [1.0 / n * (2 * i - 1) for i in self.y_input]
        return tuple(costs_0), tuple(costs_1), self.X_prime_0

    def get_baseline(self, y, predictions):
        """Return the baseline FP or FN rate of the classifier predictions.

        Args:
            :param y: true labels (binary)
            :param predictions: predictions of classifier (soft predictions)

        Returns:
            :return: The baseline FP or FN rate of the classifier predictions
        """
        if self.fairness_def == 'FP':
            return np.mean([predictions[i] for i, c in enumerate(y) if c == 0])
        elif self.fairness_def == 'FN':
            return np.mean([(1 - predictions[i]) for i, c in enumerate(y)
                            if c == 1])

    def update_costs(self, c_0, c_1, group, C, iteration, gamma):
        """Recursively update the costs from incorrectly predicting 1 for the learner.

        Args:
            :param c_0: current costs for predicting 0
            :param c_1: current costs for predicting 1
            :param group: last group found by the auditor, object of class Group.
            :param C: see Model class for details.
            :param iteration: current iteration
            :param gamma: target disparity

        Returns:
            :return c_0, c_1: tuples of new costs for CSC problem of learner
        """

        # make costs mutable type
        c_0 = list(c_0)
        c_1 = list(c_1)

        pos_neg = group.disparity_direction
        n = len(self.y)

        g_members = group.func.predict(self.X_prime_0)
        m = self.X_prime_0.shape[0]
        g_weight = np.sum(g_members) * (1.0 / float(m))
        for i in range(n):
            X_prime_0_index = 0
            if self.y[i] == 0:
                new_group_cost = (1.0 / n) * pos_neg * C * (
                    1.0 / iteration) * (g_weight - g_members[X_prime_0_index])
                if np.abs(group.weighted_disparity) < gamma:
                    new_group_cost = 0

                if self.fairness_def == 'FP':
                    c_1[i] = (c_1[i] - 1.0 / n) * (
                        (iteration - 1.0) /
                        iteration) + new_group_cost + 1.0 / n
                elif self.fairness_def == 'FN':
                    c_0[i] = (c_0[i] - 1.0 / n) * (
                        (iteration - 1.0) /
                        iteration) + new_group_cost + 1.0 / n

                X_prime_0_index += 1
            else:
                if self.fairness_def == 'FP':
                    c_1[i] = -1.0 / n
                elif self.fairness_def == 'FN':
                    c_0[i] = -1.0 / n
        return tuple(c_0), tuple(c_1)

    def get_subset(self, predictions):
        """Returns subset of dataset with y = 0 for FP and labels, or subset with y = 0 with flipped labels
        if the fairness_def is FN.

        Args:
            :param predictions: soft predictions of the classifier
        Returns:
            :return: X_prime_0: subset of features with y = 0
            :return: labels: the labels on y = 0 if FP else 1-labels.
        """
        if self.fairness_def == 'FP':
            return self.X_prime_0, [
                a for u, a in enumerate(predictions) if self.y[u] == 0
            ]
        # handles FN rate by flipping labels
        elif self.fairness_def == 'FN':
            return self.X_prime_0, [(1 - a) for u, a in enumerate(predictions)
                                    if self.y[u] == 0]

    def get_group(self, predictions, metric_baseline):
        """Given decisions on sensitive attributes, labels, and FP rate audit wrt
            to gamma unfairness. Return the group found, the gamma unfairness, fp disparity, and sign(fp disparity).

        Args:
            :param predictions: soft predictions of the classifier
            :param metric_baseline: see function get_baseline

        Returns:
            :return func: object of type RegOracle defining the group
            :return g_size_0: the size of the group divided by n
            :return fp_disp: |group_rate-baseline|
            :return fp_disp_w: fp_disp*group_size_0
            :return sgn(fp_disp): sgn(group_rate-baseline)
            :return fp_group_rate_neg:
        """

        X_subset, predictions_subset = self.get_subset(predictions)

        m = len(predictions_subset)
        n = float(len(self.y))

        cost_0 = [0.0] * m
        cost_1 = -1.0 / n * (metric_baseline - predictions_subset)

        reg0 = linear_model.LinearRegression()
        reg0.fit(X_subset, cost_0)
        reg1 = linear_model.LinearRegression()
        reg1.fit(X_subset, cost_1)
        func = RegOracle(reg0, reg1)
        group_members_0 = func.predict(X_subset)

        # get the false positive rate in group
        if sum(group_members_0) == 0:
            fp_group_rate = 0
        else:
            fp_group_rate = np.mean([
                r for t, r in enumerate(predictions_subset)
                if group_members_0[t] == 1
            ])
        g_size_0 = np.sum(group_members_0) * 1.0 / n
        fp_disp = np.abs(fp_group_rate - metric_baseline)
        fp_disp_w = fp_disp * g_size_0

        cost_0_neg = [0.0] * m
        cost_1_neg = -1.0 / n * (predictions_subset - metric_baseline)

        reg0_neg = linear_model.LinearRegression()
        reg0_neg.fit(X_subset, cost_0_neg)
        reg1_neg = linear_model.LinearRegression()
        reg1_neg.fit(X_subset, cost_1_neg)
        func_neg = RegOracle(reg0_neg, reg1_neg)
        group_members_0_neg = func_neg.predict(X_subset)

        if sum(group_members_0_neg) == 0:
            fp_group_rate_neg = 0
        else:
            fp_group_rate_neg = np.mean([
                r for t, r in enumerate(predictions_subset)
                if group_members_0[t] == 0
            ])
        g_size_0_neg = np.sum(group_members_0_neg) * 1.0 / n
        fp_disp_neg = np.abs(fp_group_rate_neg - metric_baseline)
        fp_disp_w_neg = fp_disp_neg * g_size_0_neg

        # return group
        if (fp_disp_w_neg > fp_disp_w):
            return Group(func_neg, g_size_0_neg, fp_disp_w_neg, fp_disp_neg,
                         -1, fp_group_rate)
        else:
            return Group(func, g_size_0, fp_disp_w, fp_disp, 1,
                         fp_group_rate_neg)

    def audit(self, predictions):
        """Takes in predictions on dataset (X',y) and returns:
            a membership vector which represents the group that violates the fairness metric,
            along with the gamma disparity.
        """
        if isinstance(predictions, pd.DataFrame):
            predictions = predictions.values

        metric_baseline = self.get_baseline(self.y_input, predictions)
        group = self.get_group(predictions, metric_baseline)

        return group.func.predict(self.X_prime), group.weighted_disparity



# -*- coding: utf-8 -*-
#
# aif360-bias-toolbox documentation build configuration file, created by
# sphinx-quickstart on Fri Jun  8 14:05:59 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

from datetime import datetime
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))

import aif360
import fairlearn


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax']

flv = tuple(map(int, fairlearn.__version__.split('.')))
if flv > (0, 7, 0) and flv[-1] == 0:
    flv = flv[:-1]
flv = '.'.join(map(str, flv))
intersphinx_mapping = {
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'sklearn': ('https://scikit-learn.org/stable/', None),
    'fairlearn': (f'https://fairlearn.org/v{flv}/', None),
    'python': ('https://docs.python.org/{}.{}'.format(*sys.version_info), None),
    'inFairness': ('https://ibm.github.io/inFairness/', None),
    'skorch': ('https://skorch.readthedocs.io/en/stable/', None),
}

napoleon_include_init_with_doc = True
napoleon_use_ivar = True
napoleon_use_rtype = False

autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['templates']

# generate autosummary even if no references
autosummary_generate = True

autodoc_default_flags = [
    'members',
    # 'inherited-members'
]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
# source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'aif360'
copyright = f'2018 - {datetime.now().year}, The AI Fairness 360 (AIF360) Authors'
author = 'aif360 developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = aif360.__version__
# The short X.Y version.
version = '.'.join(release.split('.')[:2])

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['templates']

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = 'literal'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# If true, parameters with trailing underscores will be properly escaped.
strip_signature_backslash = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['static']

def setup(app):
    app.add_css_file('style.css')

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',  # needs 'show_related': True theme option to display
#         'searchbox.html',
#         'donate.html',
#     ]
# }


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'aif360doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'aif360.tex', u'AI Fairness 360 Documentation',
     u'IBM Research', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'aif360', u'AI Fairness 360 Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'aif360', u'AI Fairness 360 Documentation',
     author, 'aif360', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for linkcode -------------------------------------------------
# taken from numpy/doc/source/conf.py:
import inspect
from os.path import relpath, dirname
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != 'py':
        return None

    modname = info['module']
    fullname = info['fullname']

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split('.'):
        try:
            obj = getattr(obj, part)
        except Exception:
            return None

    # strip decorators, which would resolve to the source of the decorator
    # possibly an upstream bug in getsourcefile, bpo-1764286
    try:
        unwrap = inspect.unwrap
    except AttributeError:
        pass
    else:
        obj = unwrap(obj)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    if lineno:
        linespec = "#L%d-L%d" % (lineno, lineno + len(source) - 1)
    else:
        linespec = ""

    fn = relpath(fn, start=dirname(aif360.__file__))

    return "https://github.com/Trusted-AI/AIF360/blob/master/aif360/%s%s" % (
           fn, linespec)
